# This module has been set as an imsi entry point, and is used by SSS
# However, it also contains functionality that replaces the set_env_vars functions in shell_functions
# and overlaps / builds from the old cylc get-time-vars.
#
# What should be done is to split apart the part of this that is SSS specific, and put that in
# SSS, which should be made it own stand alone tool within imsi
#
# The, the part that is more generally useful (write_env_variables_to_file), should be made
# its own tool, and add to as needed to replace the shell_functions currently used.
# Part of the expansion / refactor of this and time_manager, should be to use these tools
# to do the setting etc of both the internal loop (replace shell loops / searches) and
# the external loop setting in both maestro and cylc.
import click
import csv
import json
import sys
from imsi.tools.time_manager.cftime_utils import _parse_iso8601_with_reso, get_date_type
from imsi.tools.time_manager.time_manager import SimDuration, SimTimer
from datetime import timedelta

"""
Module for managing chunk progression based on tracking and chunk files.

This module provides functionality to read time chunk definitions from a CSV file,
track the current position of a timer in a JSON tracking file, and manage
incrementing through defined time chunks. It can also produce an file containing "standard"
CCCma timer variables, that can be sourced in the shell.

It includes a command-line interface
(CLI) for easy interaction.
"""

def read_chunks(file_path):
    """Read the chunk file and return the chunks as a list of lists.

    Each chunk is represented by [RESTART, START, END], where these values are
    timestamps in a string format.

    Parameters:
    - file_path (str): Path to the CSV file containing the time chunks.

    Returns:
    - chunks (list): A list of lists, each representing a time chunk.
    """
    try:
        with open(file_path, 'r') as f:
            reader = csv.reader(f)
            chunks = [row for row in reader]
        return chunks
    except FileNotFoundError:
        click.echo(f"Error: The file {file_path} does not exist.", err=True)
        sys.exit(1)

def load_tracking_data(tracking_file):
    """Load the tracking data from the JSON tracking file.

    Parameters:
    - tracking_file (str): Path to the tracking file.

    Returns:
    - data (dict): The tracking data as a dictionary. Will raise:
       - FileNotFoundError if file does not exist
       - json.JSONDecodeError if file is mangled
    """
    try:
        with open(tracking_file, 'r') as f:
            return json.load(f)
    except FileNotFoundError as e:
        raise e
    except json.JSONDecodeError as e:
        raise e

def get_tracking_data(tracking_file):
    """Get the tracking data from the JSON tracking file.

    Parameters:
    - tracking_file (str): Path to the tracking file.

    Returns:
    - data (dict): The tracking data as a dictionary. If no file exists,
    return an empty dict.
    """
    try:
        return load_tracking_data(tracking_file)
    except FileNotFoundError:
        return {}

def save_tracking_data(tracking_file, data):
    """Save the tracking data to the JSON tracking file.

    Parameters:
    - tracking_file (str): Path to the tracking file.
    - data (dict): Dictionary to save in the tracking file.
    """
    with open(tracking_file, 'w') as f:
        json.dump(data, f, indent=4)

def get_current_chunk(tracking_file, timer_name):
    """Get the current tracked chunk for the specified timer from the tracking file.

    Parameters:
    - tracking_file (str): Path to the tracking file.
    - timer_name (str): The name of the timer for which the current chunk is being tracked.

    Returns:
    - current_chunk (dict): The current chunk as a dictionary (RESTART, START, END), or None if not found.
    """
    data = load_tracking_data(tracking_file)
    return data.get(timer_name)

def set_current_chunk(tracking_file, timer_name, restart, start, end):
    """Write the full chunk (RESTART, START, END) to the tracking file under the specified timer_name.

    Parameters:
    - tracking_file (str): Path to the tracking file.
    - timer_name (str): The name of the timer to track the chunk for.
    - restart (str): The RESTART time of the chunk.
    - start (str): The START time of the chunk.
    - end (str): The END time of the chunk.
    """
    data = get_tracking_data(tracking_file)
    data[timer_name] = {
        "RESTART": restart,
        "START": start,
        "END": end
    }
    save_tracking_data(tracking_file, data)

def get_chunk_by_start(chunks, start_time):
    """Find a chunk by its START time.

    Parameters:
    - chunks (list): List of time chunks.
    - start_time (str): The START time to search for.

    Returns:
    - chunk (list): The found chunk, or None if not found.
    - index (int): The index of the found chunk, or -1 if not found.
    """
    for idx, chunk in enumerate(chunks):
        if chunk[1] == start_time:
            return chunk, idx
    return None, -1

def increment_chunk_logic(chunk_file, tracking_file, timer_name):
    """
    Move to the next time chunk for a given timer and update the tracking file.

    This function locates the current chunk for the specified timer, increments
    to the next chunk in the sequence, and updates the tracking file accordingly.

    Outputs status codes for shell usage:
    - 0: success (the chunk was incremented)
    - 1: end_of_chunks (no more chunks available)
    - 2: initialized_first_chunk (first chunk being tracked)

    Parameters:
    - chunk_file (str): Path to the file containing the list of available time chunks.
    - tracking_file (str): Path to the tracking file where the current chunk status is stored.
    - timer_name (str): The name of the timer to track the chunks for.
    """
    chunks = read_chunks(chunk_file)
    current_chunk = get_current_chunk(tracking_file, timer_name)

    if current_chunk:
        # Find the index of the current chunk
        current_start = current_chunk["START"]
        _, current_idx = get_chunk_by_start(chunks, current_start)

        # Get the next chunk if available
        next_idx = current_idx + 1
        if next_idx < len(chunks):
            next_chunk = chunks[next_idx]
            set_current_chunk(tracking_file, timer_name, next_chunk[0], next_chunk[1], next_chunk[2])
            return "0"  # success
        else:
            return "1"  # end_of_chunks
    else:
        # Default to the first chunk if no tracking data exists
        first_chunk = chunks[0]
        set_current_chunk(tracking_file, timer_name, first_chunk[0], first_chunk[1], first_chunk[2])
        return "2"  # initialized_first_chunk


def init_tracking_file(chunk_file, tracking_file, timer_name, overwrite_timer=False):
    """
    Initialize the tracking file with the first chunk for a specified timer.

    Parameters:
    - chunk_file (str): Path to the CSV file containing the time chunks.
    - tracking_file (str): Path to the tracking file where the current chunk status is stored.
    - timer_name (str): The name of the timer to track.

    If the timer_name already exists in the tracking file, no changes will be made.
    """
    chunks = read_chunks(chunk_file)
    first_chunk = chunks[0]

    # Load existing tracking data
    data = get_tracking_data(tracking_file)

    # Only add if the timer does not already exist or overwrite is on (True)
    if (timer_name not in data) or overwrite_timer:
        data[timer_name] = {
            "RESTART": first_chunk[0],
            "START": first_chunk[1],
            "END": first_chunk[2]
        }
        save_tracking_data(tracking_file, data)
    else:
        print(f"Timer '{timer_name}' is already initialized.")

def parse_cftime(date_string, calendar):
        """Parse ISO 8601 date string into a cftime object."""
        parsed, _ = _parse_iso8601_with_reso(get_date_type(calendar), date_string)
        return parsed

def segment_chunk_size_to_cftime(start_date_str, stop_date_str, chunk_size, loop_index, calendar):
    """
    Calculate current dates from segment dates, chunk size, and loop index.
    Return in cftime object format.

    :param start_date_str: Date string in format YYYY-MM-DDTHH:MM:SS (ISO 8601)
    :param stop_dat_str: Date string in format YYYY-MM-DDTHH:MM:SS (ISO 8601)
    :param chunk_size: Duration string in format PnYnMnDTnHnMnS (ISO 8601)
    :param loop_index: Integer index of the desired chunk
    :param calendar: Calendar system for cftime (e.g., "gregorian", "noleap", "360_day")
    """

    sim_timer = SimTimer.from_iso('job', start_date_str, stop_date_str, chunk_size,
                      'job', 'job', calendar=calendar)
    current_start_date = sim_timer.ChunkIndexStart[loop_index-1]
    current_stop_date = sim_timer.ChunkIndexStop[loop_index-1]
    current_restart_date = current_start_date - timedelta(seconds=1)

    return current_restart_date, current_start_date, current_stop_date

def date_strings_to_cftime(restart_date_str, start_date_str, stop_date_str, calendar="gregorian"):
    """
    Take current time strings and convert to cftime objects.

    The restart, start, and stop dates are expected in ISO 8601 format but are processed using `cftime`
    to handle a wide range of calendar systems.

    :param restart_date_str: Date string in format YYYY-MM-DDTHH:MM:SS (ISO 8601)
    :param start_date_str: Date string in format YYYY-MM-DDTHH:MM:SS (ISO 8601)
    :param stop_date_str: Date string in format YYYY-MM-DDTHH:MM:SS (ISO 8601)
    :param calendar: Calendar system for cftime (e.g., "gregorian", "noleap", "360_day")
    """

    start_date_cftime = parse_cftime(start_date_str, calendar)
    stop_date_cftime = parse_cftime(stop_date_str, calendar)
    restart_date_cftime = parse_cftime(restart_date_str, calendar)

    return restart_date_cftime, start_date_cftime, stop_date_cftime

def date_start_duration_to_cftime(start_date_str, duration_str, calendar="gregorian"):
    """
    Take start time and duration strings and convert to cftime objects.

    The date and duration are expected in ISO 8601 format but are processed using `cftime`
    to handle a wide range of calendar systems.

    :param start_date_str: Date string in format YYYY-MM-DDTHH:MM:SS (ISO 8601)
    :param duration_str: Duration string in format PnYnMnDTnHnMnS (ISO 8601)
    :param calendar: Calendar system for cftime (e.g., "gregorian", "noleap", "360_day")
    """

    start_date_cftime, reso = _parse_iso8601_with_reso(get_date_type(calendar), start_date_str)
    duration = SimDuration(duration_str)

    stop_date_cftime = start_date_cftime + duration.duration - timedelta(seconds=1)
    restart_date_cftime = start_date_cftime - timedelta(seconds=1)

    return restart_date_cftime, start_date_cftime, stop_date_cftime

def check_if_first_chunk(timer_name, chunk_file, tracking_file):
    # compare time in timer tracking file with FIRST entry of list in
    # chunk file for the given timer

    # chunk file:
    chunk_times = read_chunks(chunk_file)
    first_chunk_times = chunk_times[0]

    # timer tracking file:
    current_timer = get_current_chunk(tracking_file, timer_name)

    is_first_chunk = list(current_timer.values()) == first_chunk_times
    return is_first_chunk

def write_env_variables_to_file(restart_date, start_date, stop_date, prefix, output_file):
    """
    Write environment variables based on restart, start, and stop dates into a shell script.
    These variables can be sourced in a shell session.
    The restart, start, and stop dates are expected in ISO 8601 format but are processed using `cftime`
    to handle a wide range of calendar systems.

    :param restart_date: Date as a cftime object
    :param start_date: Date as a cftime object
    :param stop_date: Date as a cftime object
    :param prefix: String prefix for environment variables
    :param output_file: Path to output shell file
    :param calendar: Calendar system for cftime (e.g., "gregorian", "noleap", "360_day")
    """

    def extract_cftime_fields(cftime_obj):
        """Extract year, month, day, and hour fields from a cftime object."""
        return {
            "year": cftime_obj.year,
            "month": cftime_obj.month,
            "day": cftime_obj.day,
            "hour": cftime_obj.hour,
            "YYYY": f"{cftime_obj.year:04d}",
            "MM": f"{cftime_obj.month:02d}",
            "DD": f"{cftime_obj.day:02d}",
            "HH": f"{cftime_obj.hour:02d}"
        }

    # Parse the restart, start, and stop dates using cftime
    restart_info = extract_cftime_fields(restart_date)
    start_info = extract_cftime_fields(start_date)
    stop_info = extract_cftime_fields(stop_date)

    # Create a list to hold all the environment variable strings
    env_vars = [
        f'export {prefix}_restart_date="{restart_date.isoformat()}"',
        f"export {prefix}_restart_year={restart_info['year']}",
        f"export {prefix}_restart_month={restart_info['month']}",
        f"export {prefix}_restart_day={restart_info['day']}",
        f"export {prefix}_restart_hour={restart_info['hour']}",
        f"export {prefix}_restart_YYYY={restart_info['YYYY']}",
        f"export {prefix}_restart_MM={restart_info['MM']}",
        f"export {prefix}_restart_DD={restart_info['DD']}",
        f"export {prefix}_restart_HH={restart_info['HH']}",

        f'export {prefix}_start_date="{start_date.isoformat()}"',
        f"export {prefix}_start_year={start_info['year']}",
        f"export {prefix}_start_month={start_info['month']}",
        f"export {prefix}_start_day={start_info['day']}",
        f"export {prefix}_start_hour={start_info['hour']}",
        f"export {prefix}_start_YYYY={start_info['YYYY']}",
        f"export {prefix}_start_MM={start_info['MM']}",
        f"export {prefix}_start_DD={start_info['DD']}",
        f"export {prefix}_start_HH={start_info['HH']}",

        f'export {prefix}_stop_date="{stop_date.isoformat()}"',
        f"export {prefix}_stop_year={stop_info['year']}",
        f"export {prefix}_stop_month={stop_info['month']}",
        f"export {prefix}_stop_day={stop_info['day']}",
        f"export {prefix}_stop_hour={stop_info['hour']}",
        f"export {prefix}_stop_YYYY={stop_info['YYYY']}",
        f"export {prefix}_stop_MM={stop_info['MM']}",
        f"export {prefix}_stop_DD={stop_info['DD']}",
        f"export {prefix}_stop_HH={stop_info['HH']}"
    ]

    # Write all the environment variables to the file in one go
    with open(output_file, "w") as f:
        f.write("\n".join(env_vars) + "\n")



import subprocess
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import List, Dict, Optional

# TODO:
# - deal with submission prefix
#        if "submit_prefix" in self.sim.machine_config['batch_commands'].keys():
#            self.submission_command =  self.sim.machine_config['batch_commands']["submit_prefix"] + " " + self.submission_command

@dataclass
class BatchJob:
    '''Defines a batch job with specific directives'''
    user_script: List[str]
    job_directives: Dict[str, str] = field(default_factory=dict)

    def construct_job_script(self, scheduler: 'Scheduler') -> List[str]:
        '''Constructs the entire job script from the header and user script using the scheduler's directives'''

        # A bit of a messy hack. We could insist that shebangs are provided separately....
        if self.user_script and self.user_script[0].startswith('#!'):
            shebang = self.user_script[0]
            script_body = self.user_script[1:]
        else:
            shebang = ''
            script_body = self.user_script

        job_header = scheduler.construct_job_header(self.job_directives)

        # Combine the shebang, job header, and script body (space before)
        job_script = [shebang] if shebang else []
        job_script.extend(job_header)
        job_script.extend([''] + script_body)

        return job_script

@dataclass
class Scheduler(ABC):
    '''Defines an abstract interface for scheduler classes'''
    name: str
    directive_prefix: str
    submission_command: str
    queue_info_command: str
    cancel_command: str
    output_redirect: str

    @abstractmethod
    def construct_job_header(self, job_directives: List) -> List:
        '''Constructs the job header based on directives'''
        pass

    @abstractmethod
    def submit(self, job: BatchJob) -> str:
        '''Submits a job and returns the job ID'''
        pass

    @abstractmethod
    def query_queue(self) -> str:
        '''Queries the job queue and returns the queue status'''
        pass

    @abstractmethod
    def cancel_job(self, job_id: str) -> str:
        '''Cancels a job given its job ID'''
        pass

    def _run_command(self, command: List[str]) -> str:
        '''Helper method to run a command and return its output'''
        result = subprocess.run(command, capture_output=True, text=True)
        result.check_returncode()
        return result.stdout.strip()

@dataclass
class PBSScheduler(Scheduler):
    name: str = "PBS"
    directive_prefix: str = '#PBS'
    submission_command: str = 'qsub'
    queue_info_command: str = 'qstat'
    cancel_command: str = 'qdel'
    output_redirect: str = '-o {PATH}.o'
    default_directives: list = field(default_factory=lambda: ["-l walltime=02:00:00", "-l select=1:mem=100GB", "-j oe"])

    def construct_job_header(self, job_directives: List) -> str:
        directives = [f"{self.directive_prefix} {directive}" for directive in job_directives]
        return directives

    def submit(self, job_script_filename: str) -> str:
        command = [self.submission_command, job_script_filename]
        output = self._run_command(command)
        job_id = output.split('.')[0]  # Extract the job ID
        return job_id

    def query_queue(self) -> str:
        command = [self.queue_info_command]
        return self._run_command(command)

    def cancel_job(self, job_id: str) -> str:
        command = [self.cancel_command, job_id]
        return self._run_command(command)

@dataclass
class SLURMScheduler(Scheduler):
    name: str = "SLURM"
    directive_prefix: str = '#SBATCH'
    submission_command: str = 'sbatch'
    queue_info_command: str = 'squeue'
    cancel_command: str = 'scancel'
    output_redirect: str = '-o {PATH}_%j.out'
    default_directives: list = field(default_factory=lambda: ["--time=06:00:00","--nodes=1"])

    def construct_job_header(self, job_directives: List) -> str:
        directives = [f"{self.directive_prefix} {directive}" for directive in job_directives]
        return directives

    def submit(self, job_script_filename: str) -> str:
        command = [self.submission_command, job_script_filename]
        output = self._run_command(command)
        job_id = output.split()[-1]  # Extract the job ID
        return job_id

    def query_queue(self) -> str:
        command = [self.queue_info_command]
        return self._run_command(command)

    def cancel_job(self, job_id: str) -> str:
        command = [self.cancel_command, job_id]
        return self._run_command(command)

def create_scheduler(scheduler_name: str) -> Scheduler:
    '''Factory method to create a scheduler instance based on the scheduler name'''
    if scheduler_name.lower() == 'pbs':
        return PBSScheduler()
    elif scheduler_name.lower() == 'slurm':
        return SLURMScheduler()
    else:
        raise ValueError(f"Unknown scheduler name: {scheduler_name}")

# Example usage (out of date)
def main():
    user_script_content = "echo 'Hello, World!'"

    pbs_scheduler = create_scheduler('pbs')
    pbs_job = Job(user_script=user_script_content, job_directives={'nodes': '1', 'walltime': '01:00:00', 'account': 'my_account', 'queue': 'my_queue'})

    print("PBS Job Script:")
    print(pbs_job.construct_job_script(pbs_scheduler))

    job_id = pbs_scheduler.submit(pbs_job)
    print(f"PBS Job ID: {job_id}")

    queue_status = pbs_scheduler.query_queue()
    print("PBS Queue Status:")
    print(queue_status)

    cancel_output = pbs_scheduler.cancel_job(job_id)
    print("PBS Cancel Output:")
    print(cancel_output)

    slurm_scheduler = create_scheduler('slurm')
    slurm_job = Job(user_script=user_script_content, job_directives={'nodes': '1', 'walltime': '01:00:00', 'account': 'my_account', 'queue': 'my_partition'})

    print("SLURM Job Script:")
    print(slurm_job.construct_job_script(slurm_scheduler))

    job_id = slurm_scheduler.submit(slurm_job)
    print(f"SLURM Job ID: {job_id}")

    queue_status = slurm_scheduler.query_queue()
    print("SLURM Queue Status:")
    print(queue_status)

    cancel_output = slurm_scheduler.cancel_job(job_id)
    print("SLURM Cancel Output:")
    print(cancel_output)

if __name__ == "__main__":
    main()



def parse_config_inheritance(configs, selected_config, config_hierarchy=None):
    """
    Parse a configuration recursively inheriting attributes from all "parents".

    Inputs:
    -------
       configs: dict
          dictionary of all possible configurations, typically from an imsi_database.
       selected_config: str
          The configuration to choose
       config_hierarchy: list
          Typically not user specified, but used in the recursive function call to
          layer configurations in the correct order of inheritance.
    """
    if config_hierarchy is None:
        config_hierarchy = []

    config = configs[selected_config]
    config_hierarchy.append(config)

    if 'inherits_from' in config and config['inherits_from']:
        parent_configs = config['inherits_from']
        if isinstance(parent_configs, str):
            parent_configs = [parent_configs]

        for parent_config in reversed(parent_configs):
            if parent_config in configs.keys():
                config = parse_config_inheritance(configs=configs, selected_config=parent_config, config_hierarchy=config_hierarchy)
            else:
                raise NameError(f'{parent_config} not found. Check inheritance in {selected_config}')
    else:
        if len(config_hierarchy) > 1:
            config = {}
            config_hierarchy.reverse()
            for child in config_hierarchy:
                config = update(config, child)

    return config

def update(config, child):
    """
    Update the config with values from child, overriding existing keys.

    Inputs:
    -------
       config: dict
          The configuration to be updated.
       child: dict
          The configuration with updates.

    Returns:
    --------
       dict: Updated configuration.
    """
    for key, value in child.items():
        if isinstance(value, dict) and key in config:
            config[key] = update(config[key], value)
        else:
            config[key] = value
    return config

# Example usage
configs = {
    'base_config': {
        'param1': 'value1',
        'param2': 'value2'
    },
    'config1': {
        'inherits_from': 'base_config',
        'param2': 'override_value2',
        'param3': 'value3'
    },
    'config2': {
        'inherits_from': ['config1', 'config3'],
        'param4': 'value4'
    },
    'config3': {
        'inherits_from': '',
        'param5': 'value5'
    }

}

selected_config = 'config2'
resolved_config = parse_config_inheritance(configs, selected_config)
print(resolved_config)


from imsi.tools.ensemble.config import (
    load_config,
)

import pytest
from unittest.mock import MagicMock, patch
from omegaconf import OmegaConf

from pathlib import Path

parent_dir = Path(__file__).resolve().parent

@pytest.fixture
def mock_cfg():
    """Mock configuration with default settings."""
    return MagicMock(
        ensemble_level=OmegaConf.create({
            "user": "test_user",
            "config_table": "holder",
        }),
        member_level=OmegaConf.create({}),
    )


def test_load_config_csv(mock_cfg):
    """Test loading a CSV configuration."""
    with patch("imsi.tools.ensemble.table_utils.table_model.CSVTableVars") as mock_csv:
        mock_csv.return_value.get_table.return_value = "mock_csv_table"
        mock_cfg.ensemble_level["config_table"] = (
            str(parent_dir / "test_table_utils" / "test_config.csv")
        )
        ensemble_config, table = load_config(mock_cfg, show_diffs=False)
        assert table == [
                {'setup': {'runid': 'run-01', 'exp': 'cmip6-piControl'}},
                {'setup': {'runid': 'run-02', 'exp': 'cmip6-piControl'}}
        ]


def test_load_config_text(mock_cfg):
    """Test loading a CSV configuration."""
    with patch("imsi.tools.ensemble.table_utils.table_model.TextTableVars") as mock_csv:
        mock_csv.return_value.get_table.return_value = "mock_csv_table"
        mock_cfg.ensemble_level["config_table"] = (
            str(parent_dir / "test_table_utils" / "test_config.txt")
        )
        ensemble_config, table = load_config(mock_cfg, show_diffs=False)
        assert table == [
                {'setup': {'runid': 'run-01', 'exp': 'cmip6-piControl'}},
                {'setup': {'runid': 'run-02', 'exp': 'cmip6-piControl'}}
        ]


def test_load_config_unsupported_file_type(mock_cfg):
    """Test loading an unsupported file type."""
    mock_cfg.ensemble_level["config_table"] = (
        str(parent_dir / "test_table_utils" / "test_config_unsupported_file_type.db")
    )

    with pytest.raises(ValueError, match="Unsupported file type:"):
        load_config(mock_cfg, show_diffs=False)



import copy
import collections
import os
import textwrap
from typing import List

from imsi.config_manager.config_manager import Configuration, Machine
from imsi.utils.dict_tools import flatten

# FUTURE: consider
# - a template in the imsi input jsons rather than (implicitly) constructing a
# template through code below.
# - moving some steps up to config_manager to resolve config

def generate_computational_environment(comp_env: dict, machine_name, compiler_name: str = None) -> List[str]:
    """Generate content of a file that can be sourced in a shell
       to set the computational environment for compiling and running the model.

       Here we just generate and return a list of strings, that will be written
       to file elsewhere using a common tool.
    """
    comp_env_content = list()

    # Note: order of keys is important (Ordereddict might be necessary) because they
    # represent the order of shell vars/commands
    comp_env_content.append("# Imsi created model environment file for compiling and running\n")
    comp_env_content.append(f"# Created for {machine_name}\n")

    # modules
    #

    # The below code combines the module actions under "all" with those under the specific
    # compiler being used and writes these out to the file.
    # It is important NOT to change the order of the keys below, as module command
    # order is important.
    # ---
    # TODO This is definitely a "configuration" step, not an interface step!
    if compiler_name in comp_env["modules"]:
        module_config_compiler = comp_env["modules"][compiler_name]
    else:
        module_config_compiler = {}

    module_config = comp_env["modules"]["all"]
    if module_config:
        # Surely there is a better way than this V!
        # Combine the lists of keys, strictly preserving the order of the
        # all keys list. Normally I would use a set() if order was not important.
        compiler_keys = list(module_config_compiler.keys())
        all_keys = list(module_config.keys())
        for k in all_keys:
            if k in compiler_keys:
                compiler_keys.remove(k)
        module_cmd_keys = all_keys + compiler_keys
        # Here we append the strings together.
        # Note that this is purely additive.
        if module_cmd_keys:
            comp_env_content.append("\n# Module definitions")
        for modcmd in module_cmd_keys:
            argstr = ""
            if modcmd in module_config_compiler.keys():
                argstr += " ".join(module_config_compiler[modcmd])
            if modcmd in module_config.keys():
                argstr += " " + " ".join(module_config[modcmd])
            comp_env_content.append(f"module {modcmd} {argstr}")

    #
    # environment variables
    #
    comp_env_content.append("\n# Environment variables")
    env_variable_config = comp_env["environment_variables"]["all"]
    for k, v in env_variable_config.items():
        comp_env_content.append(f"export {k}={v}")

    #
    # environment commands
    #
    comp_env_content.append("\n # Environment commands")
    env_command_config = comp_env["environment_commands"]["all"]

    # script contents
    for k, v in env_command_config.items():
        comp_env_content.append(f"{k} {v}")

    return comp_env_content


def generate_computational_environment_controller(
        machine_config: Machine, run_config_path: str, basename_prefix: str = None) -> List[str]:
    """Generate the contents of the computational environment 'controller' file.

    The controller is a passthrough shell file that simply sources the correct
    computational_environment file for the machine at runtime. It does so
    by matching the hostname of the current machine (via regex) to the machine
    name (used in imsi config). Both the regex and machine name are known from
    upstream config and supplied via the `machine_config` (instance of `Machine`).
    The file that is sourced is then: `{run_config_path}/{basename_prefix}_{machine name}`.

    Parameters:
        machine_config: Machine
            machine configuration object
        run_config_path: str
            path to run config folder (usually `{runid}/config`)
        basename_prefix: str
            file basename prefix for file names to source.

    Returns:
        contents of the controller file, a list of strings (lines)
    """
    if basename_prefix is None:
        basename_prefix = "computational_environment_"

    # map machine name -> host name regex
    machine_name = machine_config.name
    machine_host_lookup = {machine_name: machine_config.nodename_regex}
    machine_host_lookup |= {k:v['nodename_regex'] for k,v in machine_config.site.items()}

    # regex transform for for shell case statement
    machine_host_lookup_shell = {k:v.replace('.*', "*") for k,v in machine_host_lookup.items()}

    # case block - options
    CASE_OPTION_TEMPLATE = """{cs})\n    COMP_ENV_MAC="{machine}"\n    ;;"""
    case_options = [
        textwrap.indent(CASE_OPTION_TEMPLATE.format(cs=r, machine=m), " " * 4)
        for m, r in machine_host_lookup_shell.items()
    ]

    # case block - full
    CASE_BLOCK_TEMPLATE_l = (
        ["case $HOSTNAME in"]
        + case_options
        + ['    *)', '        >&2 echo "Unknown machine for $HOSTNAME"', "exit 1", "        ;;", "esac;"]
    )

    # full script
    filename_template = os.path.join(run_config_path, f'{basename_prefix}$COMP_ENV_MAC')
    script_contents = (
        ['#!/bin/bash', 'HOSTNAME=$( hostname )']
        + CASE_BLOCK_TEMPLATE_l
        + [f'source {filename_template}']
    )

    return script_contents

def generate_compilation_template(configuration: Configuration):
    """Generate the compilation_template file contents based on a Configuraiton instance

    Here we just generate and return a list of strings, that will be written
    to file elsewhere using a common tool.
    """
    machine_name = configuration.machine.name
    compiler_name = configuration.compiler.name
    compiler_config = copy.deepcopy(configuration.compiler.model_dump())
    comp_env_content = list()

    comp_env_content.append("# Imsi created model environment file for compilation\n")
    comp_env_content.append(f"# Created for the compiler: {compiler_name} on machine: {machine_name}\n")
    comp_env_content.append("# Compiler flag options")
    for language, lang_config in compiler_config.items():
        comp_env_content.append(f'\n#{language} options')
        if isinstance(lang_config, collections.abc.MutableMapping):
            flat_lang_config = flatten(lang_config)
            for k,v in flat_lang_config.items():
                if isinstance(v, list):
                    argstr=' '.join(v)
                else:
                    argstr=v
                comp_env_content.append(f'{k}={argstr}')
    return comp_env_content

if __name__ == "__main__":
    pass


from imsi.config_manager.schema.post_processing import PostProcessing


def test_minimal_config():
    """Test minimal configuration of Compiler."""

    arbitrary_config = {
        'name': 'post_processing',
        'compiler': {
            'name': 'post_processing',
            'asd_executable': 'post_processing',
            'banana_flags': ['flag1', 'flag2'],
            'arghhhh_matey!': ['option1', 'option2'],
        },
    }

    # test that any config works here!
    PostProcessing(**arbitrary_config)



import pytest
from pydantic import ValidationError
from unittest.mock import MagicMock
from imsi.config_manager.databases import ConfigDatabase
from imsi.config_manager.schema.machine import MachineFactory, Machine, BatchCommands


@pytest.fixture
def valid_machine_data():
    """Fixture to provide valid machine data."""
    return {
        'name': 'test_machine',
        'site': {},
        'nodename_regex': 'test.*',
        'supported_compilers': ['gcc', 'intel'],
        'default_compiler': 'gcc',
        'scratch_dir': '/scratch',
        'storage_dir': '/storage',
        'sequencers': ['sequencer1', 'sequencer2'],
        'default_sequencing_suffix': '_seq',
        'batch_commands': {
            'scheduler': 'slurm',
            'mpi_exec': 'mpirun',
            'pes_per_node': 16,
        },
        'login_nodes': ['node1', 'node2'],
        'modules': {'python': '3.8'},
        'environment_variables': {'OMP_NUM_THREADS': '4'},
        'environment_commands': {'source': '/env/setup.sh'},
        'resources': {'cpu': 64, 'memory': '128GB'},
        'setup': None
    }


@pytest.fixture
def valid_batch_commands():
    """Fixture to provide valid batch commands."""
    return {
        'scheduler': 'slurm',
        'mpi_exec': 'mpirun',
        'pes_per_node': 16,
    }


def test_missing_required_fields():
    """Ensure required fields cannot be omitted."""
    with pytest.raises(ValidationError):
        Machine(site={})  # Missing required fields


def test_invalid_default_compiler(valid_machine_data):
    """Ensure default_compiler must be in supported_compilers."""
    valid_machine_data['default_compiler'] = 'nonexistent_compiler'
    with pytest.raises(ValueError, match='not in supported compilers'):
        Machine(**valid_machine_data)


def test_optional_fields(valid_machine_data):
    """Ensure optional fields can be omitted or set to None."""
    valid_machine_data.pop('login_nodes')
    valid_machine_data['modules'] = None
    valid_machine_data['environment_variables'] = None
    valid_machine_data['environment_commands'] = None
    valid_machine_data['resources'] = {'1': '2'}
    valid_machine_data.pop('setup')
    machine = Machine(**valid_machine_data)
    assert machine.modules is None
    assert machine.login_nodes is None  # Should default to None
    assert machine.environment_variables is None
    assert machine.environment_commands is None
    assert machine.setup is None


def test_missing_resources(valid_machine_data):
    """Ensure resources are defined in the site configuration or machine configuration."""
    valid_machine_data['resources'] = None
    valid_machine_data['site'] = {'resources': None}
    with pytest.raises(ValueError, match='Resources must be defined'):
        Machine(**valid_machine_data)


def test_invalid_nodename_regex(valid_machine_data):
    """Ensure nodename_regex is a valid regular expression."""
    valid_machine_data['nodename_regex'] = '[a-z'
    with pytest.raises(
        ValueError, match='Invalid regular expression for nodename_regex'
    ):
        Machine(**valid_machine_data)


def test_create_from_database():
    """Ensure create_from_database correctly retrieves machine info."""
    mock_db = MagicMock(spec=ConfigDatabase)
    mock_db.get_parsed_config.side_effect = lambda section, name: {
        'nodename_regex': 'test.*',
        'supported_compilers': ['gcc', 'intel'],
        'default_compiler': 'gcc',
        'scratch_dir': '/scratch',
        'storage_dir': '/storage',
        'sequencers': ['seq1'],
        'default_sequencing_suffix': '_seq',
        'batch_commands': {'scheduler': 'slurm', 'mpi_exec': 'mpirun'},
        'resources': {'cpu': 64, 'memory': '128GB'},
    }
    mock_db.get_config.side_effect = lambda section: {
        'sites': {'test_site': ['test_machine']},
        'machines': {'test_machine': {'nodename_regex': 'test.*'}},
    }

    _, machine = Machine.create_from_database(mock_db, 'test_machine')
    assert machine.name == 'test_machine'
    assert machine.default_compiler == 'gcc'


@pytest.mark.parametrize('required_key', ['scheduler', 'mpi_exec'])
def test_batch_commands_missing(valid_batch_commands, required_key):
    valid_batch_commands.pop(required_key)
    with pytest.raises(ValidationError):
        BatchCommands(**valid_batch_commands)


def test_batch_commands_optional_fields(valid_batch_commands):
    """Ensure optional fields can be omitted or set to None."""
    valid_batch_commands['pes_per_node'] = None
    batch_commands = BatchCommands(**valid_batch_commands)
    assert batch_commands.pes_per_node is None
    assert batch_commands.pes_per_node is None  # Should default to None
    assert batch_commands.mpi_exec == 'mpirun'
    assert batch_commands.scheduler == 'slurm'


def test_validate_compiler(valid_machine_data):
    """Ensure validate_compiler raises error if compiler is not in supported_compilers."""
    valid_machine_data['supported_compilers'] = ['gcc', 'intel']
    machine = Machine(**valid_machine_data)
    with pytest.raises(ValueError, match=r'Compiler .* is not supported by machine .*'):
        machine.validate_compiler('nonexistent_compiler')


def test_find_matching_machine(valid_machine_data):
    machine_config_dict = {valid_machine_data['name']: Machine(**valid_machine_data)}
    with pytest.raises((NameError, AttributeError, TypeError)):
        MachineFactory.find_matching_machine(machine_config_dict)



import click

@click.command(
    short_help="Deletes a run's directories.",
    help="""Deletes a run's directories.

    Default (no flags): contents of scratch_dir and storage_dir are deleted.
    """
    )
@click.option(
    "--runid_path",
    required=True,
    type=click.Path(),
    help="Path to a valid run setup directory.",
)
@click.option("-s", "--setup/--no-setup", default=False,
              help="Clean setup directory.", show_default=True)
@click.option("-t", "--scratch-dir/--no-scratch-dir", "temp", default=True,
              help="Clean scratch data directory.", show_default=True)
@click.option("-d", "--storage-dir/--no-storage-dir", "data", default=True,
              help="Clean storage data directory.", show_default=True)
@click.option("-a", "--all", "clean_all", is_flag=True, help="Clean all locations.")
def clean(runid_path, setup, temp, data, clean_all):
    # If no options selected by default cleans data and temp directories
    from imsi.tools.disk_tools.disk_tools import clean_run

    if clean_all:
        setup, temp, data = (True, True, True)

    if not any((setup, temp, data)):
        click.echo("Nothing to clean")
    else:
        clean_run(runid_path, setup, temp, data)



from . import shell_interface_utilities as shell_interface_utilities
from . import shell_config_parameters as shell_config_parameters
from . import shell_comp_environment as shell_comp_environment
from . import shell_inputs_outputs as shell_inputs_outputs
from . import shell_diag_parameters as shell_diag_parameters
from . import shell_timing_vars as shell_timing_vars



# tests/test_ensemble/test_comprehensive/test_cli.py
from pathlib import Path
import shutil

import pytest
from click.testing import CliRunner

from imsi.tools.ensemble.ensemble_cli import ensemble


CONFIG_DIR = Path.cwd() / "tests" / "test_ensemble" / "test_comprehensive"  # directory containing this file
CONFIG_FILES = sorted(CONFIG_DIR.glob("config*.yaml"))
RUN_DIRECTORY = Path.cwd() / "tests" / "test_ensemble" / "test_comprehensive" / "tmp_setup_dirs"  # directory for test runs

# Skip the whole module cleanly if nothing to test
if not CONFIG_FILES:
    pytest.skip(
        f"No configuration files matching 'config*.yaml' found in {CONFIG_DIR}.",
        allow_module_level=True,
    )

@pytest.fixture(scope="session")
def runner() -> CliRunner:
    """A single CliRunner reused for every parametrised test."""
    return CliRunner()


@pytest.fixture
def setup_dir(tmp_path):
    """
    A unique temporary directory for each test invocation,
    e.g. /tmp/pytest-of-<user>/pytest-NNNN/p0.
    If your CLI needs to *see* this directory (via an env-var or flag),
    monkeypatch or add an option here.
    """
    return tmp_path


@pytest.mark.filterwarnings("ignore:\\x1b\\[[0-9;]*m?Overlapping keys:UserWarning")
@pytest.mark.parametrize(
    "config_file",
    CONFIG_FILES,
    ids=[p.name for p in CONFIG_FILES],  # pretty node names in `pytest -vv`
)
def test_setup_and_save_restarts(runner, setup_dir, config_file):
    """Ensure `setup` succeeds for every config YAML."""
    RUN_DIRECTORY.mkdir(parents=True, exist_ok=True)
    try:
        result = runner.invoke(
            ensemble,
            ["--config-path", str(config_file), "setup"],
        )
    except FileNotFoundError as e:
        print(f"File not found: {e}. Expected error if in GitLab CI environment.")
    finally:
        shutil.rmtree(RUN_DIRECTORY, ignore_errors=True)



import pytest
from imsi.tools.ensemble.table_utils.table_utils import (
    get_keys,
    remove_explicit_none,
    convert_to_bracket_notation,
    replace_aliases,
    add_nested_key,
    warn_on_overlapping_keys,
)
from imsi.tools.ensemble.config import load_config

from omegaconf import OmegaConf, DictConfig
import copy
from pathlib import Path

parent_dir = Path(__file__).resolve().parent


def make_cfg(mapping) -> DictConfig:
    """Convenience wrapper to create DictConfig objects from plain dicts."""
    return OmegaConf.create(mapping)


@pytest.fixture
def config_params(request):
    pars = {
        "repo": request.config.getoption("--target-repo"),
        "branch": request.config.getoption("--target-branch"),
        "model": request.config.getoption("--target-model"),
        "experiment": request.config.getoption("--target-exp")
    }
    return pars


@pytest.fixture
def setup_config(config_params):
    return make_cfg({
        "ver": config_params["branch"],
        "repo": config_params["repo"],
        "fetch_method": "clone",
        "machine": "ppp6",
        "model": config_params["model"],
        "runid": ["run1"],
        "exp": config_params["experiment"]
    })


@pytest.fixture
def config_data_one(setup_config):
    """Configuration data for the first test case."""
    return make_cfg({
        "member_level": {
            "setup": setup_config,
        }
    })


@pytest.fixture
def config_data_two(setup_config):
    """Configuration data for the second test case."""
    return make_cfg({
        "member_level": {
            "setup": {
                **setup_config,
                "model": [setup_config["model"], setup_config["model"]],  # override for test
                "runid": ["run-yaml-001", "run-yaml-002"],
            }
        }
    })


@pytest.mark.filterwarnings("ignore:\\x1b\\[[0-9;]*m?Overlapping keys:UserWarning")
def test_get_keys(config_data_two):
    cfg = make_cfg({
        "ensemble_level": {
            "user": "test_user",
            "run_directory": str(parent_dir / "."),
            "config_table": str(parent_dir / "config_table.yaml"),
        },
        "member_level": config_data_two["member_level"],
    })

    ensemble_config, constructed_table = load_config(cfg, show_diffs=False)
    assert get_keys(constructed_table[1]) == {
        'setup:exp',
        'components',
        'components:CanAM',
        'components:CanAM:namelists:canam_settings',
        'setup:runid',
        'components:CanAM:namelists:canam_settings:pp_rdm_num_pert',
        'setup:repo',
        'setup',
        'components:CanAM:namelists',
        'setup:ver',
        'setup:machine',
        'setup:fetch_method',
        'setup:model'
    }


@pytest.fixture
def config_with_explicit_none():
    """Configuration with an explicit None value."""
    return {
        "setup": {
            "runid": "run-yaml-002",
            "exp": "test_exp",
        },
        "components": {
            "CanAM": {"namelists": {"canam_settings": {"pp_rdm_num_pert": None}}}
        }
    }


@pytest.fixture
def config_with_explicit_none_removed():
    """Configuration with explicit None values removed."""
    return {
        "setup": {
            "runid": "run-yaml-002",
            "exp": "test_exp",
        },
        "components": {"CanAM": {"namelists": {"canam_settings": {}}}}
    }


def test_remove_explicit_none(config_with_explicit_none, config_with_explicit_none_removed):
    expected = config_with_explicit_none_removed
    assert remove_explicit_none(config_with_explicit_none) == expected


@pytest.mark.parametrize(
    "raw, expected",
    [
        ("setup:runid",        "setup[runid]"),
        ("root:child:0",       "root[child][0]"),
        ("singlekey",          "singlekey"),          # no delimiter → unchanged
        ("a:b:c:d",            "a[b][c][d]"),
        ("numeric:123:4",      "numeric[123][4]"),    # numbers stay as strings
        ("leading:",           "leading[]"),          # trailing empty segment
        (":trailing",          "[trailing]"),         # leading empty segment
        ("",                   ""),                   # empty string remains empty
    ],
)
def test_convert_to_bracket_notation_happy(raw, expected):
    assert convert_to_bracket_notation(raw) == expected


@pytest.mark.parametrize(
    "bad_input, exception",
    [
        (None,              AttributeError),   # .split fails on NoneType
        (123,               AttributeError),   # int has no .split
        (["a:b"],           AttributeError),   # list has no .split
    ],
)
def test_convert_to_bracket_notation_type_errors(bad_input, exception):
    with pytest.raises(exception):
        convert_to_bracket_notation(bad_input)


def test_single_alias_replaced():
    """
    Top-level key 'exp' is moved under 'setup.exp', and the original is removed.
    """
    cfg = make_cfg({"exp": "piControl"})
    aliases = {"exp": "setup:exp"}
    result = replace_aliases(cfg, aliases)

    assert "exp" not in result                    # alias key removed
    assert result.setup.exp == "piControl"        # value moved
    # same DictConfig instance mutated in-place:
    assert result is cfg


def test_alias_absent_no_change():
    original = make_cfg({"foo": 1})
    cfg_copy = copy.deepcopy(original)
    aliases = {"missing": "setup:missing"}
    result = replace_aliases(original, aliases)

    assert result == cfg_copy                    # deep equality
    assert result is original                    # same object


def test_multiple_aliases():
    cfg = make_cfg({"exp": "piControl", "runid": ["01", "02"]})
    aliases = {"exp": "setup:exp", "runid": "setup:runid"}

    replace_aliases(cfg, aliases)

    assert OmegaConf.to_container(cfg) == {
        "setup": {"exp": "piControl", "runid": ["01", "02"]}
    }


def test_merge_into_existing_path():
    """
    If 'setup' already exists, new keys are merged (not overwritten) because
    replace_aliases calls OmegaConf.update(..., merge=True).
    """
    cfg = make_cfg(
        {
            "setup": {"existing": 123},
            "exp": "amip",
            "runid": "03",
        }
    )
    aliases = {"exp": "setup:exp", "runid": "setup:runid"}

    replace_aliases(cfg, aliases)

    # Existing content is preserved; new keys merged
    assert OmegaConf.to_container(cfg) == {
        "setup": {"existing": 123, "exp": "amip", "runid": "03"}
    }


def test_add_top_level_key():
    cfg = make_cfg({})
    add_nested_key(cfg, "foo", 123)

    assert cfg.foo == 123
    assert OmegaConf.to_container(cfg) == {"foo": 123}


def test_add_nested_key():
    cfg = make_cfg({})
    add_nested_key(cfg, "setup:exp", "piControl")

    assert cfg.setup.exp == "piControl"
    assert OmegaConf.to_container(cfg) == {"setup": {"exp": "piControl"}}


def test_merge_into_existing_node():
    cfg = make_cfg({"setup": {"existing": 1}})
    add_nested_key(cfg, "setup:exp", "amip")

    assert cfg.setup.existing == 1            # untouched existing data
    assert cfg.setup.exp == "amip"
    assert OmegaConf.to_container(cfg) == {
        "setup": {"existing": 1, "exp": "amip"}
    }


def test_overlap_emits_warning():
    """
    Both configs contain 'setup:runid' so a UserWarning must be raised.
    """
    cfg1 = make_cfg({"setup": {"runid": "01"}})
    cfg2 = make_cfg({"setup": {"runid": "02", "other": 1}})

    # `pytest.warns` passes when *exactly one* warning of the requested type
    # occurs (additional warnings are also captured/allowed).
    with pytest.warns(UserWarning) as record:
        warn_on_overlapping_keys(cfg1, cfg2)

    # Optional: assert the message contains the overlapping key
    assert "setup:runid" in str(record[0].message)


def test_no_overlap_emits_no_warning(recwarn):
    """
    Distinct key sets: the function should be silent.
    """
    cfg1 = make_cfg({"setup": {"runid": "01"}})
    cfg2 = make_cfg({"model": {"name": "canam"}})

    warn_on_overlapping_keys(cfg1, cfg2)

    # `recwarn` is the warnings recorded list; should stay empty
    assert len(recwarn) == 0




import pytest
import json
import os
from click.testing import CliRunner
from imsi.tools.disk_tools import disk_tools
from imsi.cli.entry import cli
from pathlib import Path
import subprocess
import shutil
import pickle
from omegaconf import OmegaConf

@pytest.fixture
def runner():
    return CliRunner()


# sets up working directory for tests with data, temp_data, and a minimal setup directory
def setup_clean_test(fail_type=None):
    cwd = Path.cwd()

    # create empty data and temp data directories
    data_dir = cwd.joinpath("data")
    data_dir.mkdir()
    temp_dir = cwd.joinpath("temp_data")
    temp_dir.mkdir()
    data_runid = data_dir.joinpath("imsi-test")
    data_runid.mkdir(parents=True)

    temp_runid = temp_dir.joinpath("maestro", "imsi-test")
    if fail_type != "zero":  # Makes it so there should be 0 runids in temp location
        temp_runid.mkdir(parents=True)

    if fail_type == "multi":  # makes there be two run_ids in the same location
        temp_runid2 = temp_dir.joinpath("imsi-test")
        temp_runid2.mkdir(parents=True)

    # create setup directory with config file and src to make clean work
    config_dir = cwd.joinpath("imsi-test")
    config_dir.mkdir()

    # mock .imsi state dir
    imsi_config_state = config_dir / ".imsi"
    imsi_config_state.mkdir()

    src = config_dir.joinpath("src")
    src.mkdir()

    config_file = config_dir.joinpath("imsi_configuration_imsi-test.yaml")
    script_location = Path(__file__).parent.resolve()
    source_config = f"{script_location}/imsi_configuration_imsi-test.yaml"

    config_data = OmegaConf.load(source_config)

    config_data["machine"]["storage_dir"] = str(data_dir)
    config_data["machine"]["scratch_dir"] = str(temp_dir)

    OmegaConf.save(config_data, config_file)

    # also pickle the file into .imsi
    with open(imsi_config_state / '.imsi_configuration_imsi-test.pickle', 'wb') as file:
        pickle.dump(config_data, file)

    return config_dir, data_runid, temp_runid


@pytest.mark.parametrize(
    ("args", "directories"),
    [
    ([],  [True, False, False]),
    (["-s", "-t", "-d"], [False, False, False]),
    (["-s"], [False, False, False]),
    ]
)
def test_clean_run(runner, args, directories):
    with runner.isolated_filesystem():
        setup_dir, data_dir, temp_dir = setup_clean_test()
        result = runner.invoke(cli, ["clean", "--runid_path=imsi-test"] + args)
        assert result.exit_code == 0, result.output
        assert setup_dir.exists() == directories[0], "setup directory exists when it should have been deleted"
        assert temp_dir.exists() == directories[1], "temp directory exists when it should have been deleted"
        assert data_dir.exists() == directories[2], "data directory exists when it should have been deleted"


def test_multi_runids(runner):
    with runner.isolated_filesystem():
        _ = setup_clean_test("multi")
        with pytest.raises(ValueError):
            runner.invoke(cli,
                        ["clean", "--runid_path=imsi-test", "-a"],
                        catch_exceptions=False
                        )

def test_wrk_dir_set(runner):
    try:
        os.environ["WRK_DIR"] = "abc/123/doremi/"
        with runner.isolated_filesystem():
            _ = setup_clean_test()
            result = runner.invoke(cli, ["clean", "--runid_path=imsi-test"])
            assert result.exit_code != 0, result.output
    finally:
        os.environ["WRK_DIR"] = ""



from importlib.resources import files
import logging
from pathlib import Path
import subprocess
import sys

import imsi
from imsi.utils.general import get_active_venv
from imsi.user_interface.setup_manager import ValidatedSetupOptions


IMSI_STATEFUL_FOLDERS = ['src', 'config']

_logger_config = {
    'setup': {
        'logger_name': 'imsi-setup',
        'filename': '.imsi-setup.log',
        'hidden': True,
        'level': logging.DEBUG
    },
    'cli': {
        'logger_name': 'imsi-cli',
        'filename': '.imsi-cli.log',
        'hidden': True,
        'level': logging.INFO
    },
    'runtime': {
        'logger_name': 'imsi-runtime',
        'filename': '.imsi-cli.log',
        'hidden': True,
        'level': logging.INFO
    }
}


def _get_logger(logger_name: str=None, path: Path=None, filename: str=None,
                hidden=True, level=logging.INFO):
    # Return a logging.Logger with the name `logger_name` set at logging
    # level `level`. The log file will be written to the full path
    # constructed from `path / filename`. The `filename` must end with
    # extension `.log`. The `hidden=True` flag ensures that the filename
    # begins with a period (`.`) (conversely, ensures it is not included
    # if `hidden=False`).

    path = Path(path)

    if not path.exists():
        raise FileNotFoundError(f"ERROR: can't write log file to path {path}")
    if not filename.endswith('.log'):
        raise ValueError('filename must end with .log')

    logger = logging.getLogger(logger_name)
    if logger.hasHandlers():
        logger.handlers.clear()

    if hidden:
        basename = f".{filename}" if not filename.startswith('.') else filename
    else:
        basename = filename.lstrip('.')

    formatter = logging.Formatter(
        fmt='%(asctime)s - %(thread)d - %(name)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%dT%H:%M:%S%z'
        )

    # instantiate log config
    fh = logging.FileHandler(str(path / basename))
    fh.setFormatter(formatter)
    logger.addHandler(fh)
    logger.setLevel(level)

    return logger


def get_imsi_logger(name, path):
    """Return an imsi logger of name `name` written to the log file `path`."""
    logger_settings = _logger_config[name]
    logger_settings['path'] = path
    return _get_logger(**logger_settings)


def log_setup(cli_args, setup_params: ValidatedSetupOptions, with_src_status=True):
    """Log setup command and information to the setup log file."""
    setup_path = Path(setup_params.runid).resolve()

    logger = get_imsi_logger('setup', setup_path)
    logger.info(f"🚀 IMSI setup for {setup_params.runid} 🚀")
    logger.info('setup_params: {}'.format(' '.join(f"{k}={v}" for k, v in setup_params.model_dump().items())))

    if with_src_status:
        # capture the status of the git repo under /src
        s = files("imsi").joinpath("utils/repo_query_status.sh")
        src_path = setup_path / 'src'
        try:
            proc = subprocess.run([str(s)], capture_output=True, cwd=src_path)
            proc.check_returncode()
        except subprocess.CalledProcessError as e:
            src_msg = f'Could not determine status under {src_path}'
        else:
            src_msg = f'src: {proc.stdout.decode().strip()}'

        logger.info(src_msg)

    # DEV: keep this as the last setup log entry (newline)
    input_args = ' '.join(sys.argv)
    logger.info(f"Setup command used:\n{input_args}")


def imsi_log_prelude(func_name, logger):
    """Log header information (imsi metadata) to the `logger` related to
    running `func_name`. Logging is set to logging.INFO.
    """
    imsi_meta = f"imsi {imsi.__version__} {imsi.__path__[0]}"
    imsi_venv = get_active_venv()
    logger.info(f'INVOKING {func_name}')
    logger.info(' '.join(sys.argv))
    logger.info(imsi_meta)
    logger.info(f'VIRTUAL_ENV {imsi_venv}')


def imsi_log_postlude(func_name, logger):
    """Log footer information related to running `func_name`.
    Logging level is set to logging.INFO.
    """
    logger.info(f'COMPLETED {func_name}')


def imsi_state_snapshot(path, folders=None, logger=None):
    """Take a snapshot of the imsi state and return the state hash.

    This runs the snapshot tool (cli_snapshot_state.sh) for the `folders`
    specified under the `path`.

    Parameters:
        path : path to an imsi run folder
        folders : folder names (relative to `path`) on which to take the state
            snapshot. These folders must be git repos. Default list is
            ['config', 'src'].
        logger : instance of a Logger.

    Returns:
        the hash of the imsi state folder
    """

    if folders is None:
        folders = IMSI_STATEFUL_FOLDERS
    elif isinstance(folders, str):
        folders = [folders]

    tracker_script = Path(__file__).parent / 'cli_snapshot_state.sh'
    folder_args = [l for s in [['-p', p] for p in folders] for l in s]
    io_args = ['-i', path, '-o', path]

    if not tracker_script.exists():
        raise FileNotFoundError(f'ERROR missing snapshot tool: {tracker_script}')
    try:
        proc = subprocess.run([str(tracker_script)] + folder_args + io_args, capture_output=True)
        proc.check_returncode()
    except subprocess.CalledProcessError as e:
        if logger:
            logger.error(f'state snapshot failed')
            logger.error(f'HALTING')
        raise ChildProcessError("Failed call: {cmd}\n{err}".format(cmd=' '.join(proc.args), err=proc.stdout.decode())) from e

    state_sha = proc.stdout.decode().strip()
    if logger:
        logger.info(f'IMSI state:{state_sha}')

    return state_sha



import pandas as pd
from pathlib import Path
from unittest.mock import patch, MagicMock
from imsi.tools.list.list_manager import (
    _filter_dataframe,
    _load_model_experiment_pairs,
    gather_choices_df,
)


def test_filter_dataframe():
    df = pd.DataFrame({
        "model": ["A", "B", "A"],
        "experiment": ["exp1", "exp2", "exp2"],
    })
    filters = {"model": "A", "experiment": None}
    filtered = _filter_dataframe(df, filters)
    assert all(filtered["model"] == "A")


def test_load_model_experiment_pairs():
    mock_db = MagicMock()
    mock_db.get_config.return_value = ["exp1"]
    mock_db.get_parsed_config.return_value = {"supported_models": ["A", "B"]}

    with patch("imsi.tools.list.list_manager.database_factory", return_value=mock_db):
        df = _load_model_experiment_pairs(Path("fake"))
    assert set(df["model"]) == {"A", "B"}
    assert set(df["experiment"]) == {"exp1"}


def test_gather_choices_df_filters(tmp_path):
    repo = tmp_path
    config_path = Path("imsi-config")

    mock_df = pd.DataFrame({
        "model": ["A", "B"],
        "experiment": ["exp1", "exp2"],
    })

    with patch("imsi.tools.list.list_manager._load_model_experiment_pairs", return_value=mock_df):
        df = gather_choices_df(
            repo_paths=[repo],
            repo_sources=["src"],
            relative_imsi_config_path=config_path,
            filter_model="A"
        )
    assert set(df["model"]) == {"A"}
    assert set(df["experiment"]) == {"exp1"}



import click
from pathlib import Path
from dotenv import dotenv_values
import os
from importlib.resources import files


def load_rc(variable=None, default=None):
    """
    Load environment variables from the site and user-specific .env files.
    This function is called at the start of the CLI to ensure that all necessary
    environment variables are available for the commands.
    """

    config_site = dotenv_values(dotenv_path=files("imsi").joinpath("imsi.site.rc"))
    config_user = dotenv_values(dotenv_path=f"{os.getenv('HOME')}/imsi.user.rc")
    config_env = os.environ

    var_list = []
    for config, source in zip(
        (config_site, config_user, config_env),
        ("imsi.site.rc", "imsi.user.rc", "$IMSI_DEFAULT_CONFIG_REPOS"),
    ):
        value = config.get(variable, None)
        if value is None:
            continue

        # Always split on colon
        # if nones exist (give [value])
        for v in str(value).split(":"):
            if v:  # ignore empty parts from "::"
                var_list.append((v, source))

    return var_list or None


def get_repo_paths(repo_path: str | None) -> tuple[list[Path], Path]:
    """
    Resolves and validates repository paths containing 'imsi-config' directories.

    Depending on the provided `repo_path` argument and the current working directory,
    this function determines which repositories to use, validates their configuration
    requirements, and returns a list of valid repository paths along with the relative
    path to the 'imsi-config' directory.

    Args:
        repo_path (str | None): Optional path to a specific repository. If None, the function
            will attempt to use the current working directory or load default repositories
            from the runtime configuration.

    Returns:
        tuple[list[Path], Path]: A tuple containing:
            - A list of Path objects, each representing a repository containing an 'imsi-config' directory.
            - A Path object representing the relative path to the 'imsi-config' directory.

    Raises:
        ValueError: If no valid repositories with an 'imsi-config' directory are found,
            or if any provided path is not a directory.
    """
    from imsi.user_interface.ui_manager import validate_version_reqs

    run_dir_config = Path.cwd() / "src" / "imsi-config"

    if not repo_path and run_dir_config.exists():
        validate_version_reqs(run_dir_config)
        return [Path.cwd()], ["pwd"], Path("src", "imsi-config")

    if repo_path:
        base_paths = [(Path(repo_path), "CLI args")]
    else:
        rc_paths = load_rc("IMSI_DEFAULT_CONFIG_REPOS")
        if not rc_paths:
            raise ValueError("No repositories provided via --repo-path or IMSI_DEFAULT_CONFIG_REPOS in user or site rc file.")
        base_paths = [(Path(p), source) for p, source in rc_paths]

    paths, sources = [], []
    for base, source in base_paths:
        if not base.is_dir():
            raise ValueError(f"Path '{base}' is not a directory.")

        if (base / "imsi-config").is_dir():
            matches = [base]
        else:
            matches = [p for p in base.iterdir() if p.is_dir() and (p / "imsi-config").is_dir()]

        paths.extend(matches)
        sources.extend([source] * len(matches))

    if not paths:
        raise ValueError(f"No repositories with 'imsi-config' found in: {base_paths}")

    for path in paths:
        validate_version_reqs(path / "imsi-config")

    return paths, sources, Path("imsi-config")


@click.command(
        name="list",
        short_help="List available model/exp configurations.",
        help="List available model/exp configurations.")
@click.option('--repo-path', default=None,
              help="""Path to either a single repo or a directory of repos.
              If not provided, examines the repos listed in the .rc.""")
@click.option('--filter-model', default=None, type=str, help="Filter by model name.")
@click.option('--filter-experiment', default=None, type=str, help="Filter by experiment name.")
def list(repo_path, filter_model, filter_experiment):

    from imsi.tools.list.list_manager import list_all_choices
    paths, sources, config_path = get_repo_paths(repo_path)

    list_all_choices(
        repo_paths=paths,
        repo_sources=sources,
        relative_imsi_config_path=config_path,
        filter_model=filter_model,
        filter_experiment=filter_experiment
    )



import click
from pathlib import Path
import sys

def schema_constructor(create_func, **kwargs):
    return create_func(**kwargs)


def check_path(path: Path):
    import imsi.user_interface.ui_manager as uim

    if not path.exists() or not path.is_dir():
        raise FileNotFoundError(f"Provided path {path} does not exist or is not a directory.")

    try:
        uim.validate_version_reqs(path)
    except FileNotFoundError:
        raise SystemExit(f"the path {path} is not a valid imsi-config folder")
    except ValueError as e:
        raise e

@click.group(
    short_help="Validate imsi configuration files.",
    help="Validate imsi configuration files."
)
@click.option(
    "--imsi-config-path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    default=str(Path.cwd()),
    help="Path to the imsi-config directory.",
)
@click.pass_context
def validate(ctx, imsi_config_path):
    """A CLI tool to validate imsi configuration files"""
    from imsi.config_manager.config_manager import ConfigManager, database_factory

    # TIU still print help msg if passed if command invoked from any
    # location (otherwise will not print if check_path fails)
    if sys.argv[-1] in ctx.help_option_names:
        click.echo(ctx.command.get_command(ctx, ctx.invoked_subcommand).get_help(ctx))
        ctx.exit()

    check_path(Path(imsi_config_path))
    db = database_factory(imsi_config_path)
    cm = ConfigManager(db)
    ctx.ensure_object(dict)
    ctx.obj["cm"] = cm
    ctx.obj["imsi_config_path"] = imsi_config_path


@validate.command()
def syntax():
    """Validate the syntax of the imsi configuration files."""
    click.echo("\033[1;32m[✓] The config directory contains valid yaml.\033[0m")


@validate.command()
@click.argument("experiment", type=str)
@click.argument("model", type=str)
@click.pass_context
def experiment(ctx, experiment, model):
    """Validate an experiment configuration."""
    cm = ctx.obj["cm"]
    schema_constructor(
        cm.create_experiment,
        experiment_name=experiment,
        model_name=model,
    )

    click.echo(f"\033[1;32m[✓] Experiment {experiment} adheres to the imsi schema.\033[0m")


@validate.command()
@click.argument("model", type=str)
@click.pass_context
def model(ctx, model):
    """Validate a model configuration."""
    cm = ctx.obj["cm"]
    schema_constructor(cm.create_model, model_name=model)

    click.echo(f"\033[1;32m[✓] Model {model} adheres to the imsi schema.\033[0m")


@validate.command()
@click.argument("machine", type=str)
@click.pass_context
def machine(ctx, machine):
    """Validate a machine configuration."""
    cm = ctx.obj["cm"]
    schema_constructor(cm.create_machine, machine_name=machine)

    click.echo(f"\033[1;32m[✓] Machine {machine} adheres to the imsi schema.\033[0m")


@validate.command()
@click.argument("experiment", type=str)
@click.argument("model", type=str)
@click.pass_context
def components(ctx, experiment, model):
    """Validates an experiment configuration"""
    from imsi.utils.general import change_dir

    cm = ctx.obj['cm']
    with change_dir(ctx.obj["imsi_config_path"]):
        model_basemodel = schema_constructor(cm.create_model, model_name=model)
        experiment_basemodel = schema_constructor(
            cm.create_experiment,
            experiment_name=experiment,
            model_name=model,
        )

        schema_constructor(
            cm.create_components,
            model=model_basemodel,
            experiment=experiment_basemodel,
        )

    click.echo(f"\033[1;32m[✓] Components for experiment {experiment} and model {model} adhere to the imsi schema.\033[0m")


@validate.command()
@click.pass_context
def utilities(ctx):
    """Validate the utility configurations."""
    cm = ctx.obj["cm"]
    schema_constructor(cm.create_utilities)

    click.echo("\033[1;32m[✓] Utilities adheres to the imsi schema.\033[0m")


@validate.command()
@click.argument("sequencer_name", type=str)
@click.argument("machine_name", type=str)
@click.argument("model_name", type=str)
@click.argument("experiment_name", type=str)
@click.argument("flow_name", type=str, required=False, default="")
@click.pass_context
def sequencing(ctx, sequencer_name, machine_name, model_name, experiment_name,  flow_name):
    """Validate a sequencing configuration."""
    cm = ctx.obj["cm"]

    machine = schema_constructor(cm.create_machine, machine_name=machine_name)
    experiment = schema_constructor(
        cm.create_experiment,
        model_name=model_name,
        experiment_name=experiment_name,
    )

    schema_constructor(
        cm.create_sequencing,
        machine=machine,
        sequencer_name=sequencer_name,
        flow_name=flow_name,
        experiment=experiment
    )

    click.echo(f"\033[1;32m[✓] Sequencing {sequencer_name}, experiment {experiment_name}, model {model_name}, and machine {machine_name} adhere to the imsi schema.\033[0m")



from pydantic import BaseModel, Field, FilePath, DirectoryPath, ConfigDict
from omegaconf import DictConfig, ListConfig
from typing import Any, Dict, List, Tuple, Optional, Annotated
from imsi.tools.ensemble.table_utils.table_utils import colour, replace_aliases


def _extract_lists(
    data: Dict[str, Any], parent_keys: List[str] = None
) -> Dict[Tuple[str, ...], List[Any]]:
    """Recursively extract lists from a nested dictionary."""
    if parent_keys is None:
        parent_keys = []

    listed_vars = {}
    for key, value in data.items():
        current_keys = parent_keys + [key]
        if isinstance(value, (ListConfig, list)):
            listed_vars[tuple(current_keys)] = value
        elif isinstance(value, (DictConfig, dict)):
            # Recurse into nested dictionaries
            nested_lists = _extract_lists(value, current_keys)
            listed_vars.update(nested_lists)
    return listed_vars


def validate_list_lengths(config_dict: dict):
    """Ensure all list values in a flat dictionary have the same length."""
    list_lengths = {key: len(val) for key, val in config_dict.items()}

    if not list_lengths:
        return

    expected_len = next(iter(list_lengths.values()))
    mismatched = {k: v for k, v in list_lengths.items() if v != expected_len}

    if mismatched:
        print(colour("❌ Inconsistent list lengths found:", "red"))
        for key, length in mismatched.items():
            print(f"  {key}: length {length} (expected {expected_len})")
        raise ValueError("All list values must have the same length.")


class EnsembleLevelVars(BaseModel):
    """Class with config variables -- logically configure so that
    they only require one definition for a given ensemble run."""

    user: Annotated[str, Field(..., description="User ID")]
    run_directory: Optional[DirectoryPath] = "."
    config_table: Optional[FilePath] = None
    aliases: Optional[dict] = None
    share_repo: Optional[bool] = False


class MemberLevelVars(BaseModel):
    """Class provides main interaction with entry config data."""
    model_config: Optional[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)
    config_data: DictConfig = Field(..., description="Configuration data for the ensemble")

    @classmethod
    def get_listed_vars(cls, config_data) -> Dict[Tuple[str, ...], List[Any]]:
        """Get all lists from the configuration data, including nested ones."""
        listed_vars = _extract_lists(config_data)
        # check that all lists in the dictionary are the same length
        validate_list_lengths(listed_vars)

        return listed_vars

    @property
    def ensemble_size(self) -> int:
        """Return the size of the ensemble, assuming all lists have the same size."""
        varlist = self.get_listed_vars(self.config_data)
        return len(list(varlist.values())[0]) if varlist else None


class InputConfig(BaseModel):
    ensemble_level: EnsembleLevelVars
    member_level: MemberLevelVars

    def model_post_init(self, __context):

        listed_vars = MemberLevelVars.get_listed_vars(self.member_level.config_data)
        sizes = [
            len(value) for value in listed_vars.values() if not isinstance(value, str)
        ]
        listed_vars_size = {
            tuple(key): len(value)
            for key, value in listed_vars.items()
            if not isinstance(value, str)
        }

        if self.member_level.config_data and not self.ensemble_level.config_table:
            if not sizes:
                raise ValueError(
                    "No lists found in config. For a single ensemble member, use 'name: [var]' in the config."
                )

            if len(set(sizes)) != 1:
                size_report = {":".join(key): val for key, val in listed_vars_size.items()}

                v = {}
                for key, value in sorted(size_report.items()):
                    v.setdefault(value, []).append(key)

                raise ValueError(
                    f"All lists in the configuration data must have the same size.\n"
                    f"Sizes are {v}"
                )

        # replace aliases in member_level if they exist and are provided
        if self.ensemble_level.aliases:
            self.member_level.config_data = replace_aliases(
                self.member_level.config_data, self.ensemble_level.aliases
            )



"""Top-level package for imsi."""

__author__ = "CCCma Technical Development Team"
__email__ = ""

from importlib.metadata import version as get_version, PackageNotFoundError

try:
    # Use installed package version (when installed via pip)
    __version__ = get_version("cimsi")
except PackageNotFoundError:
    try:
        # Use locally generated setuptools_scm file (for local dev)
        from ._version import version as __version__
    except ImportError:
        __version__ = "unknown"



from imsi.tools.time_manager.time_manager import sim_time_factory
from datetime import timedelta


def generate_timers_config(run_dates: dict, timer_name: str) -> dict:
    """Generate lists of timer variables for various jobs, to be written out and used
    downstream. Specifically returns start and stop lists for the model outer loop (job submission),
    model internal loop (within jobs), and the postproc loop.


    Parameters:
        run_dates (dict ) : A dict of run dates, in iso8061 format, with keys
                            [run_, run_segment]_[start,stop]_date, and keys
                            for model_chunk_size, model_internal_chunk_size
                            and postproc_chunk_size.
        timer_name (str) : The prefix of the job timer name (e.g. model_submission_job)

        Return:
         timer_lists (dict): key=timer_name and value=list of start -- stop times
    """
    # There are two big issues here to be resolved
    # 1. There is repetition of creating SimulationTime instance, with shell_config_parameters
    # 2. This occurs because this should be set upstream in config_manager, but how precisely to
    # populate templates has not really been properly figured out yet.

    # The details of these timers might need to change (e.g. for different sequencers or even flows),
    # and others could easily be added by referring to the time_manager "_generate_job_start_end_date_lists"
    # function. In fact, those calls might really belong here.

    # Use the time_manager module to get detailed timers for insertion into shell_parameters
    simTime = sim_time_factory(run_dates, timer_name)
    cftime_start_list = getattr(simTime, 'ChunkIndexStart')
    cftime_stop_list = getattr(simTime, 'ChunkIndexStop')
    if not (cftime_start_list and cftime_stop_list):
        raise ValueError(f"The timer name: {timer_name} did not return any content.")
    timeliststr = []
    for (start_time, stop_time) in zip(cftime_start_list, cftime_stop_list):
        restart_time = start_time - timedelta(seconds=1)
        timeliststr.append(f"{restart_time.isoformat(timespec='seconds')},{start_time.isoformat(timespec='seconds')},{stop_time.isoformat(timespec='seconds')}")
    return timeliststr


def validate_timers_config(run_dates: dict):
    """
    Validate the model timers supplied by the IMSI config json.
    model_internal_chunk_size must be less than model_chunk_size
    and the resulting start and stop times must line up.
    This essentially means that all of the outer loop dates
    (model_submission_job) must be present in the list of
    inner loop dates (model_inner_loop).

    Parameters:
        run_dates (dict ) : A dict of run dates, in iso8061 format, with keys
                            [run_, run_segment]_[start,stop]_date, and keys
                            for model_chunk_size, model_internal_chunk_size
                            and postproc_chunk_size.
    """

    outer = sim_time_factory(run_dates, 'model_submission_job')
    inner = sim_time_factory(run_dates, 'model_inner_loop')

    for chunk in outer.ChunkIndexStart:
        if chunk not in inner.ChunkIndexStart:
            raise ValueError("Misaligned start dates in model_submission_job and model_inner_loop! " +
                             "Must set model_internal_chunk_size / model_chunk_size to have an integer ratio less than 1, " +
                             "and the boundaries between chunks must align.")

    for chunk in outer.ChunkIndexStop:
        if chunk not in inner.ChunkIndexStop:
            raise ValueError("Misaligned stop dates in model_submission_job and model_inner_loop! " +
                             "Must set model_internal_chunk_size / model_chunk_size to have an integer ratio less than 1, "+
                             "and the boundaries between chunks must align.")




import click
import sys
import json

@click.group()
def chunk_manager():
    """Manage time information for the simulation."""
    pass

@chunk_manager.command()
@click.option('--tracking-file', default="tracking.json", help="Path to the tracking file.")
@click.option('--timer-name', required=True, help="The name of the timer to track the chunking session.")
def current(tracking_file, timer_name):
    """Get current chunk for specific timer from chunk_file."""
    import imsi.tools.time_manager.chunk_manager as cm

    chunk_data = cm.get_current_chunk(tracking_file, timer_name)
    if chunk_data is not None:
        print(json.dumps(chunk_data, indent=2))
    else:
        click.echo(f"Error: Timer '{timer_name}' not found in tracking file.", err=True)
        sys.exit(1)

@chunk_manager.command()
@click.option('--chunk-file', required=True, type=click.Path(exists=True), help="Path to the chunks file.")
@click.option('--tracking-file', required=True, default="tracking.json", help="Path to the tracking file.")
@click.option('--timer-name', required=True, help="The name of the timer to track the chunking session.")
def increment(chunk_file, tracking_file, timer_name):
    """Move to the next time chunk for a given timer and update the tracking file."""
    import imsi.tools.time_manager.chunk_manager as cm

    status = cm.increment_chunk_logic(chunk_file, tracking_file, timer_name)
    click.echo(status)

@chunk_manager.command()
@click.option('--chunk-file', required=True, type=click.Path(exists=True), help="Path to the chunks file.")
@click.option('--tracking-file', default="tracking.json", help="Path to the tracking file.")
@click.option('--timer-name', required=True, help="The name of the timer to track the chunking session.")
def init(chunk_file, tracking_file, timer_name):
    """Initialize the tracking file with the first chunk for the specified timer."""
    import imsi.tools.time_manager.chunk_manager as cm

    cm.init_tracking_file(chunk_file, tracking_file, timer_name)

@chunk_manager.group(
    short_help="Create an environment file of the chunk's time data.",
    help='Create the environment file that contains chunk time information as environment variables.'
    )
def create_time_env_file():
    pass

@create_time_env_file.command(
    short_help="Create an environment file of the chunk's time data using the timer tracking file.",
    help="Create an environment file of the chunk's time data using the timer tracking file.",
    )
@click.option('--tracking-file', required=True, default=".simulation.time.state", type=click.Path(exists=True),
              help="Path to the timer tracking file that contains the current time information for the simulation.")
@click.option('--timer-name', required=True, help="The name of the timer to track the chunking session.")
@click.option('--prefix', required=True, help="Prefix for environment variables.")
@click.option('--output-file', required=True, type=click.Path(), help="Path to the output shell file.")
@click.option('--calendar', default="noleap", help="Calendar system for cftime (e.g., 'gregorian', 'noleap', '360_day').")
def from_tracking_file(tracking_file, timer_name, prefix, output_file, calendar):
    """
    Create a shell file with environment variables based on the current chunk data.
    The tracking file contains only the current date/time information for the simulation,
    """
    import imsi.tools.time_manager.chunk_manager as cm


    # Get the current chunk data
    chunk_data = cm.get_current_chunk(tracking_file, timer_name)
    if chunk_data is None:
        click.echo(f"Error: Timer '{timer_name}' not found in tracking file.", err=True)
        sys.exit(1)

    # Extract relevant dates from the current chunk
    restart_date = chunk_data["RESTART"]
    start_date = chunk_data["START"]
    stop_date = chunk_data["END"]

    # Convert strings into cftime objects
    restart_date_cf, start_date_cf, stop_date_cf = cm.date_strings_to_cftime(restart_date, start_date, stop_date, calendar)

    # Call the function to write the environment variables to a file
    cm.write_env_variables_to_file(restart_date_cf, start_date_cf, stop_date_cf, prefix, output_file)
    click.echo(f"Environment variables written to {output_file}.")

@create_time_env_file.command(
    short_help="Create an environment file of the chunk's time data using the chunk's time information (iso8601).",
    help="Create an environment file of the chunk's time data using the chunk's time information (iso8601)."
    )
@click.option('--isostr', type=str, required=True, help="Date string in iso8601 format representing the start date of a simulation segment")
@click.option('--chunk-delta', type=str, required=True, help="Duration string in iso8601 format")
@click.option('--prefix', required=True, help="Prefix for environment variables.")
@click.option('--output-file', required=True, type=click.Path(), help="Path to the output shell file.")
@click.option('--calendar', default="noleap", help="Calendar system for cftime (e.g., 'gregorian', 'noleap', '360_day').")
def from_date(isostr, chunk_delta, prefix, output_file, calendar):
    import imsi.tools.time_manager.chunk_manager as cm

    restart_date, start_date, stop_date = cm.date_start_duration_to_cftime(isostr, chunk_delta, calendar)

    # Call the function to write the environment variables to a file
    cm.write_env_variables_to_file(restart_date, start_date, stop_date, prefix, output_file)
    click.echo(f"Environment variables written to {output_file}.")

@create_time_env_file.command(
    short_help="Create an environment file of the chunk's time data based on the loop index.",
    help="Create an environment file of the chunk's time data based on the loop index.",
    epilog="Primarily used for the maestro sequencer."
    )
@click.option('--start-time', required=True, type=str, help="Date string in iso8601 format representing the start date of a simulation segment")
@click.option('--stop-time', required=True, type=str, help="Date string in iso8601 format represent the stop date of a simulation segment")
@click.option('--chunk-size', required=True, type=str, help="Duration string in iso8601 format representing the size of a simulation segment")
@click.option('--loop-index', required=True, type=int, help="Integer value representing the index of a simulation segment")
@click.option('--prefix', required=True, help="Prefix for environment variables.")
@click.option('--output-file', required=True, type=click.Path(), help="Path to the output shell file.")
@click.option('--calendar', default="noleap", help="Calendar system for cftime (e.g., 'gregorian', 'noleap', '360_day').")
def from_index(start_time, stop_time, chunk_size, loop_index, prefix, output_file, calendar):
    import imsi.tools.time_manager.chunk_manager as cm

    restart_date, start_date, stop_date = cm.segment_chunk_size_to_cftime(start_time, stop_time, chunk_size, loop_index, calendar)

    # Call the function to write the environment variables to a file
    cm.write_env_variables_to_file(restart_date, start_date, stop_date, prefix, output_file)
    click.echo(f"Environment variables written to {output_file}.")

@chunk_manager.command()
@click.option('--chunk-file', required=True, type=click.Path(exists=True), help="Path to the chunks file.")
@click.option('--tracking-file', default="tracking.json", help="Path to the tracking file.")
@click.option('--timer-name', required=True, help="The name of the timer to track the chunking session.")
def is_first_chunk(chunk_file, tracking_file, timer_name):
    import imsi.tools.time_manager.chunk_manager as cm

    # boolean is cast to int
    click.echo(int(cm.check_if_first_chunk(timer_name, chunk_file, tracking_file)))



import os
import glob
import getpass
from pathlib import Path
import shutil
import sys
from textwrap import dedent

from imsi.user_interface.ui_manager import load_run_config
from imsi.utils.general import change_dir, remove_folder
from imsi.cli.core_cli import force_dirs


PRINT_MESSAGE_TEMPLATE_CLEANED = "Cleaned space: {path}"

def get_assoc_paths(starting_path, runid):
    """Finds a path(s) given a runid and starting location. Returns all matches

    Max. depth is 2 to find matching folders for runid:
        starting_path/*/runid
    """
    dir_list = [Path(p) for p in glob.glob(f"{starting_path}/*/{runid}")] + [Path(p) for p in glob.glob(f"{starting_path}/{runid}")]
    return dir_list


def delete_run_folders_from_base(starting_path, runid, only_if_empty=False):
    """Deletes the runid folder from the starting location (base).

    Max. depth is 2 to find matching folders for runid:
        starting_path/*/runid

    only_if_empty: If True, the matched path directory is only
        deleted if already empty.
    """

    dirs_assoc = get_assoc_paths(starting_path, runid)

    if len(dirs_assoc) == 0:
        # no locations to delete
        return
    elif len(dirs_assoc) > 1:
        # multiple -> unclear what to delete, skip
        raise ValueError(f"Multiple subfolders with runid {runid} found under {starting_path}; skipping")
    dir_assoc_runid = dirs_assoc[0]  # safe

    if only_if_empty:
        try:
            remove_folder(dir_assoc_runid, force=False)
        except OSError:
            # if the folder is not empty, leave intact
            pass
        else:
            print(PRINT_MESSAGE_TEMPLATE_CLEANED.format(path=dir_assoc_runid))
    else:
        shutil.rmtree(dir_assoc_runid)
        print(PRINT_MESSAGE_TEMPLATE_CLEANED.format(path=dir_assoc_runid))


def _del_linked_storage_from_base(base, runid, subfolder):
    if base is not None:
        base = os.path.expandvars(base)
        base_runid = os.path.join(base, runid, subfolder)

        if os.path.exists(base_runid):
            shutil.rmtree(base_runid)
        print(PRINT_MESSAGE_TEMPLATE_CLEANED.format(path=base_runid))

        # delete the parent folder if empty only
        delete_run_folders_from_base(base, runid, only_if_empty=True)


def clean_run(runid_path, setup, temp, data):
    """Deletes folders associated with a run"""

    runid_path = Path(runid_path).resolve()
    # Check to see if WRK_DIR is set. If it is, make sure the path is the same as runid_path
    # If they are not the same we bail to prevent edge cases where you can accidentally clean
    # the wrong run
    if Path(os.getenv("WRK_DIR", runid_path)).resolve() != runid_path.resolve():
        sys.exit(dedent(f"""
            Error: 'WRK_DIR' is set to a location that differs from the runid_path provided:
              WRK_DIR: {os.getenv("WRK_DIR")}
              runid_path: {runid_path}

            unset 'WRK_DIR' before using imsi clean
              (use "unset WRK_DIR" to unset the environment variable)
            """)
        )

    runid = runid_path.name

    with change_dir(runid_path):
        # TODO DEV: force_dirs imported from cli -- slate for change of location
        # or new method needed that is built for purpose
        force_dirs(Path(f".imsi/.imsi_configuration_{runid}.pickle"))
        config = load_run_config(serialized=True)

    # Make an assumption that the run is being cleaned by the user who created the run.
    user_id = getpass.getuser()

    # deletes the information in the temp space
    if temp:
        temp_dir = config.machine.scratch_dir.replace('${USER}', user_id)
        delete_run_folders_from_base(temp_dir, runid)

    # deletes the data from the run
    if data:
        data_dir = config.machine.storage_dir.replace('${USER}', user_id)
        delete_run_folders_from_base(data_dir, runid)

    # deletes the setup directory.
    if setup:

        # deletes the dir that contains src
        # IMPORTANT: Use path from config; do *not* use realpath of src
        # as it may have been a run set up with 'link'.
        # (src_storage_dir is ignored when fetch_method=link)
        try:
            src_storage_dir = config.machine.setup.src_storage_dir
        except AttributeError:
            pass
        else:
            _del_linked_storage_from_base(src_storage_dir, runid, 'src')

        # deletes the dir that contains the exes
        try:
            exe_storage_dir = config.machine.exe_storage_dir
        except AttributeError:
            pass
        else:
            _del_linked_storage_from_base(exe_storage_dir, runid, 'bin')

        shutil.rmtree(runid_path)
        print(PRINT_MESSAGE_TEMPLATE_CLEANED.format(path=Path(runid_path).resolve()))



from imsi.utils.dict_tools import flatten

def generate_diag_parameters_content(diag_config):
    """
    This creates content for the diag_parameters file for the simulation, that contains
    variable definitions required by downstream shell scripting.
    Currently this is just a pure propagation of variables from the diagnostic
    config. Ultimately, the interaction of imsi with diagnostics needs to be refined.
    """
    diag_parameters_content = list() # to populate with strings and return

    # We are making the hard assumption that there are not duplicated keys.
    # If there are, earlier ones will be overwritten with later ones.
    flat_diag_config = flatten(diag_config, sep='>')
    flat_diag_config_onekey = {}
    for k, v in flat_diag_config.items():
        key = k.split('>')[-1]
        flat_diag_config_onekey[key] = v

    diag_parameters_content.append("# Imsi created diag environment file\n")

    for k, v in flat_diag_config_onekey.items():
        if isinstance(v, list):
            # Join list elements into a single string with spaces
            var_string = " ".join(map(str, v))
        elif isinstance(v, (str, int, float, bool)):
            # Convert non-list types to string, preserving their format
            var_string = str(v)
        else:
            # Handle unexpected types, or log an error
            raise TypeError(f"Unsupported type {type(v)} for key {k}")

        # Append the export command to the list
        diag_parameters_content.append(f'export {k}="{var_string}"')
    return diag_parameters_content



ALLOWED_JOB_STATUSES = {
    'queued': 'QUEUED',
    'running': 'RUNNING',
    'success': 'COMPLETED',
    'fail': 'FAILED'
}
ACTIVE_JOB_STATUSES = [
    ALLOWED_JOB_STATUSES['queued'],
    ALLOWED_JOB_STATUSES['running']
]

UNAWARE_DATETIME_FORMAT = "%Y%m%d%H%M%S"
UNAWARE_DATETIME_PATTERN = (
    r"(^\d{4})([0-2][0-9])([0-3][0-9])([0-2][0-9])([0-5][0-9])([0-5][0-9])"
)

JOB_STATUS_FILENAME_TEMPLATE = '{job_name}_{start_time}_{stop_time}.status'

SCRIPT_INTERPRETER_DIRECTIVE = "#!/bin/bash"
SCRIPT_COMMENT_CHAR = "#"

SCRIPT_HEADER_TEMPLATE = """
{interpreter}
set -e

# Imsi created shell environment file
# {tool_title} created on date: {date_string}
{header_comments}
{methods}

"""

SCRIPT_IMSI_ENV_COMMAND = """
# method to invoke imsi environment:
with_venv() {
    local venv_path="$1"
    shift
    ( source "$venv_path/bin/activate" && "$@" )
}
"""
# note that the body this is a raw literal string (prefix r) so that
# escapes on slashes are not needed; and it can still be used with
# f-formatted literal string
SCRIPT_BODY_TEMPLATE = r"""
job_name={job_name}
timer_name={timer_name}
timer_tracking_file={timer_tracking_file}
sequencing_dates_file={sequencing_dates_file}
sequencing_dir={sequencing_dir}
listings_dir={listings_dir}
submit_next={submit_next_bool}
submit_dependents={submit_dependents_bool}
iss_cfg={iss_config_file}
clean_scratch={clean_scratch_bool}

# source required input information regarding env vars, settings and timers.
source {run_config_dir}/computational_environment
source {run_config_dir}/shell_parameters

# Create a scratch directory
IMSI_SCRATCH={sequencing_scratch_dir}/{runid}/{runid}_scratch_{job_name}_${{{timer_name}}}_$$
mkdir -p $IMSI_SCRATCH
cd $IMSI_SCRATCH

# Parse timers for this job, and set appropriate shell variables
with_venv "$IMSI_VENV" imsi chunk-manager create-time-env-file from-tracking-file --timer-name $timer_name --tracking-file $timer_tracking_file --output-file {run_config_dir}/job_env_dates_$timer_name --prefix 'job'
source {run_config_dir}/job_env_dates_$timer_name

# Update job status
with_venv "$IMSI_VENV" imsi iss set-status -t RUNNING -j $job_name -s "$job_start_date" -e "$job_stop_date" -o "$listings_dir"

# User defined scripting
return_status=0
source {user_script} || return_status=$?

# check for errors and set status
if (( return_status == 0 )); then
    with_venv "$IMSI_VENV" imsi iss set-status -t COMPLETED -j $job_name -s "$job_start_date" -e "$job_stop_date" -o "$listings_dir"
else
    with_venv "$IMSI_VENV" imsi iss set-status -t FAILED -j $job_name -s "$job_start_date" -e "$job_stop_date" -o "$listings_dir"
    exit 1
fi

#
# wrap
#
cd {work_dir}
if $clean_scratch; then
    rm -rf $IMSI_SCRATCH
fi

#
# job cycling
#

# dependents
if [[ "$submit_dependents" = true ]]; then
    dependent_jobs=$( with_venv "$IMSI_VENV" imsi iss get-job-dependents -j $job_name -c $iss_cfg )
    for dep_job in $dependent_jobs; do
        dep_is_ready=$( with_venv "$IMSI_VENV" imsi iss ready-to-submit -j $dep_job -c $iss_cfg )
        if (( dep_is_ready == 1 )); then

            # get timer info
            d_tn=$( with_venv "$IMSI_VENV" imsi iss get-timer-name -j $dep_job -c $iss_cfg )
            d_st=$( with_venv "$IMSI_VENV" imsi iss get-timer-state -n $d_tn -i $timer_tracking_file --element START )
            d_et=$( with_venv "$IMSI_VENV" imsi iss get-timer-state -n $d_tn -i $timer_tracking_file --element END )

            # submit and update status
            with_venv "$IMSI_VENV" imsi iss submit -j $dep_job -c $iss_cfg
            with_venv "$IMSI_VENV" imsi iss set-status -t QUEUED -j $dep_job -s "$d_st" -e "$d_et" -o "$listings_dir"

        fi
    done
fi

# next iteration

# increment the timer
with_venv "$IMSI_VENV" imsi chunk-manager increment --timer-name $timer_name --tracking-file $timer_tracking_file --chunk-file $sequencing_dates_file

if [[ "$submit_next" = true ]]; then
    job_is_ready=$( with_venv "$IMSI_VENV" imsi iss ready-to-submit -j $job_name -c $iss_cfg )
    if (( job_is_ready == 1 )); then

        # get timer info
        st=$( with_venv "$IMSI_VENV" imsi iss get-timer-state -n $timer_name -i $timer_tracking_file --element START )
        et=$( with_venv "$IMSI_VENV" imsi iss get-timer-state -n $timer_name -i $timer_tracking_file --element END )

        # submit and update status
        with_venv "$IMSI_VENV" imsi iss submit -j $job_name -c $iss_cfg
        with_venv "$IMSI_VENV" imsi iss set-status -t QUEUED -j $job_name -s "$st" -e "$et" -o "$listings_dir"

    fi
fi

"""

SCRIPT_FOOTER = """
"""

# NOTE: adding these here from chunk_manager.py because they were not
# actually used in time_manager...
# these are the indices of the "columns" in a sequencing dates file.
# these are defined so that we can easily map between the date file
# and a timer file (which has a dict/json structure with key,value pairs).
# TODO: this may be changed to a class or better named data structure.
TIME_CHUNK_INDICES = {'RESTART': 0, 'START': 1, 'END': 2}



"""
IMSI Simple Sequencer (ISS)
===========================

ISS (iss) is a simple, portable batch sequencer for running jobs.

iss provides a simple time-looping mechanism to handle sequential job
submissions, and supports submission of downstream dependencies.

iss is: a set of classes and functions, a CLI (which exposes some of
the functions), and preset job script template.
"""
from collections import namedtuple
from dataclasses import dataclass, field, asdict
import datetime
from functools import cached_property
from glob import glob
import json
import os
import re
import subprocess
import textwrap
from typing import List, Dict, Optional

from imsi.scheduler_interface.schedulers import Scheduler, BatchJob
from imsi.tools.time_manager import chunk_manager as cm
from imsi.tools.time_manager.cftime_utils import parse_iso8601_like
from imsi.utils.general import get_date_string, write_shell_script
from imsi.tools.simple_sequencer.iss_globals import (
    ALLOWED_JOB_STATUSES,
    UNAWARE_DATETIME_FORMAT,
    UNAWARE_DATETIME_PATTERN,
    JOB_STATUS_FILENAME_TEMPLATE,
    SCRIPT_INTERPRETER_DIRECTIVE,
    SCRIPT_COMMENT_CHAR,
    SCRIPT_HEADER_TEMPLATE,
    SCRIPT_IMSI_ENV_COMMAND,
    SCRIPT_BODY_TEMPLATE,
    SCRIPT_FOOTER,
    TIME_CHUNK_INDICES
)


def to_shell_bool(boolean: bool):
    if boolean:
        return "true"
    else:
        return "false"

# simple containers
Parent = namedtuple('Parent', ['job_name', 'run_conditional_on'])
Dependent = namedtuple('Dependent', ['job_name', 'run_conditional_on'])

@dataclass
class SimpleJob:
    """An individal job

    Parameters:
        submission_script_fullepath (str) : Name/path of the job script.
        job_identifier (str) : Job name.
        user_script_to_source (path) :  A shell script that run's the
            job's contents (will be sourced).
        directives (list) : A list of directives for the scheduler (e.g.
            wallclock, cores).
        listings_basename (str) : Path/prefix of where to write job listings.
            [default: 'output']
        sequencing_dates_file (str) : The path to the sequencing dates file
            (CSV) which contains rows of time information for run "chunks"
            (from imsi chunk manager)
        timer_name (str) : The name of the timer for the job that will be used
            in the timer tracking file
        submit_next (bool) : If the job should submit the next iteration
            [default: True]
        submit_dependents_on_complete (bool) : if the job should submit its
            dependent jobs (listed in `depends_on`).
        depends_on (dict) : a dictionary that defines the dependent job
            names (keys) and the timer conditions for which they will be
            checked against this job (the parent) (values). See example below.
            [default: None]
        clean_scratch (bool) : Remove (delete) the scratch directory
            [default: True]

    Illustration:

        Suppose there are two jobs, `job_a` and `job_b`. `job_b` is instatiated as:

        ```
        job_b = SimpleJob(job_identifier="job_b", depends_on={"job_a": "END"}, **kwargs)
        ```

        This indicates that any iteration of `job_b` should depend on the
        previous iteration of `job_a` being successfully completed, and that
        the "END" time of `job_b` is no further along in time than the "END" of
        `job_a`.

        Do note that the `depends_on` is simply the *definition* of the
        relationship between `job_a` and `job_b`; other iss tooling can be used
        to check whether the depedency conditions are met for any job
        at any time for a particular run once set up.
    """
    # Note: these are only the ingredients that make the job, where the
    # job status is handled through run-time tooling

    submission_script_fullpath: str
    job_identifier: str
    user_script_to_source: str
    directives: list = field(default_factory=list)
    listings_basename: str ='output'
    sequencing_dates_file: str = ''
    timer_name: str = ''
    submit_next: bool = True
    submit_dependents_on_complete: bool = True
    depends_on: Optional[Dict[str, str]] = None
    clean_scratch: bool = True

def get_job_dependents(jobs: dict, job_name: str):
    """
    an ugly way to construct a relationship 'lookup' by flattening
    the config structure
    returns the information for a SINGLE job_name
    """
    if job_name not in jobs:
        raise KeyError()
    dependents = []
    for job,spec in jobs.items():
        if job == job_name:
            continue
        depends_on = spec.get('depends_on', None)
        if depends_on is not None:
            for dep in depends_on:
                if job_name == dep:
                    transposed_dep = Dependent(job, depends_on.get(job_name))
                    dependents.append(transposed_dep)
    return dependents

def get_job_parents(jobs: dict, job_name: str):
    """get parent job names for a requested job_name"""
    if job_name not in jobs:
        raise KeyError()
    r = jobs[job_name].get('depends_on', None)
    parents = []
    if r:
        for item in r.items():
            parents.append(Parent(*item))
    return parents

class SimpleFlow:
    """A simple flow object to help establish relationships between
    SimpleJob objects.

    Note: This is not akin to a formal tree/DAG and there is limited
    functionality.
    """

    def __init__(self, jobs: List[SimpleJob]):
        if isinstance(jobs, SimpleJob):
            jobs = [jobs]
        self._jobs = jobs
        self.entry = self._resolve_entry()
        self.job_dependents = self.__get_job_relationship("dependents")
        self.job_parents = self.__get_job_relationship("parents")

    @property
    def jobs(self):
        jobids = [j.job_identifier for j in self._jobs]
        if len(jobids) != len(set(jobids)):
            raise AttributeError("job identifiers must be unique")
        return {job.job_identifier: job for job in self._jobs}

    @jobs.setter
    def jobs(self, val: List[SimpleJob]):
        # FUTURE TODO dev if needed -- could do set self._jobs = val
        # and reset all other related props
        raise AttributeError("can't reset jobs once instantiated")

    @cached_property
    def __jobsdict(self):
        # convenience property
        return {job.job_identifier: asdict(job) for job in self.jobs.values()}

    def _resolve_entry(self):
        # validation
        # determine the entry job, ie. the one that doesn't have a `depends_on`
        # (which should 'None' by definition of the SimpleJob property)
        assumed_entry = [k.job_identifier for k in self._jobs if k.depends_on == None]
        if len(assumed_entry) == 1:
            # entry is defined
            self._validate_dependencies()
            return assumed_entry[0]
        elif len(self._jobs) == 1:
            # this means there is one job and it also has a depends_on
            raise AttributeError("single job flow cant have 'depends_on' attribute")
        else:
            raise AttributeError("bad flow design; can't have more than one entry job")

    def _validate_dependencies(self):
        # check properties of dependency job names (keys) and conditions (values)
        for name, j in self.__jobsdict.items():
            if j['depends_on']:
                for jn,val in j['depends_on'].items():
                    if jn == name:
                        raise KeyError(f"dependent of job '{name}' cannot be itself")
                    if jn not in self.__jobsdict:
                        raise KeyError(f"job '{name}' cannot depend on non-existent job '{jn}'")
                    if val not in TIME_CHUNK_INDICES:
                        raise ValueError(f"job dependency condition for '{name}' cannot be '{val}'")

    def __get_job_relationship(self, generation):
        # get the relationship between jobs, either the job's 'parents'
        # (upstream) or 'dependents' (downstream)
        if generation == 'parents':
            handler = get_job_parents
        elif generation == 'dependents':
            handler = get_job_dependents
        else:
            raise ValueError(
                f"handler must be one of 'parents' or 'dependents', not '{generation}'"
                )
        m = {}
        for job in self.jobs:
            m[job] = handler(self.__jobsdict, job)
        return m

@dataclass
class IMSISimpleSequencer:
    """A class to contain the top level information of a sequencer.

    Properties:
        runid : str
        work_dir: str, Path
        run_config_dir : str, Path
        sequencer_config_dir : str, Path
        sequencing_scratch_dir : str, Path
        scheduler : Scheduler
        flow : SimpleFlow
    """
    runid: str
    run_config_dir: str
    sequencer_config_dir: str
    sequencing_scratch_dir: str
    work_dir: str
    scheduler: Scheduler
    flow: SimpleFlow
    timer_tracking_file: str = field(init=False)     # required here for
    listings_dir: str = field(init=False)            #    serialization (re: asdict() vs through setter)

    def __post_init__(self):
        # just defining the path here (file created in self.configure())
        self.timer_tracking_file = os.path.join(self.sequencer_config_dir, ".simulation.time.state")
        self.listings_dir = os.path.join(self.sequencer_config_dir, "listings")

    def configure(self, force=True):
        '''Do the on disk setup of required files'''

        redirect_template = self.scheduler.output_redirect      # template

        for job_name, job in self.flow.jobs.items():

            kwargs = asdict(job)

            # create simple header
            # TODO this could be moved to SimpleJob instead
            # (could pass in more metadata from the Configuration object)
            entry_flag = to_shell_bool(self.flow.entry == job_name)
            kwargs['header_str'] = f'runid:  {self.runid}    job:  {job_name}    entry: {entry_flag}'

            # create/write the batch job using scheduler
            directives=job.directives
            if not directives:
                # TODO catch None or empty list
                directives = self.scheduler.default_directives
            output_path = os.path.join(self.listings_dir, job.listings_basename)
            output_redirect_cmd = redirect_template.format(PATH=output_path)       # not great
            directives = directives + [output_redirect_cmd]

            kwargs['output_redirect_path'] = output_redirect_cmd.split()[-1]       # not great
            script_content = self.generate_simple_shell_sequencing_content(**kwargs)

            # note: input to BatchJob must be a list
            batch_job = BatchJob(user_script=script_content.split('\n'), job_directives=directives)
            batch_script = batch_job.construct_job_script(self.scheduler)
            write_shell_script(job.submission_script_fullpath, batch_script, mode='w', make_executable=True)

            # Configure timers in timing file
            # note: this is within the loop because, though the tracking
            # file is the same, the individual timers are written to the
            # same file here
            cm.init_tracking_file(chunk_file=job.sequencing_dates_file,
                      tracking_file=self.timer_tracking_file,
                      timer_name=job.timer_name,
                      overwrite_timer=force)

    def run(self, job_name=None, init_first_job_status=True):
        """Submits the batch job to the queue.

        Optionally sets the initial status file of the job
        (`init_first_job_status=True`).
        """
        if job_name is None:
            # run the entry job if no job name provided
            job_name = self.flow.entry
        elif job_name not in self.flow.jobs:
            raise KeyError(f"can't run job '{job_name}' because it does not exist")
        job = self.flow.jobs[job_name]
        cmd_run = self.scheduler['submission_command'].split() + [job.submission_script_fullpath]
        subprocess.run(cmd_run, cwd=self.work_dir)

        # v TODO this is sort of messy because the job status is usually set
        # through the iss job scripts -- but this is the current implementation
        # to get the FIRST submission of the job to create a status file.
        # (it is NOT required for iss job scripts to work, but it is
        # ideal for the user).
        if init_first_job_status:
            job = self.flow.jobs[job_name]
            is_first = is_first_iteration(
                job_name,
                self.timer_tracking_file,
                job.sequencing_dates_file,
                timer_name=job.timer_name,
                listings_dir=self.listings_dir
                )
            if is_first:
                # set the status of job submitted
                # (timeinfo is needed to construct the name of the status file)
                timeinfo = cm.get_current_chunk(self.timer_tracking_file, job.timer_name)
                set_job_status(
                    job_name,
                    ALLOWED_JOB_STATUSES['queued'],
                    timeinfo['START'],
                    timeinfo['END'],
                    tracking_dir=self.listings_dir
                    )

        return job_name

    def generate_simple_shell_sequencing_content(
            self, job_identifier: str, user_script_to_source: str=None,
            timer_name: str='', sequencing_dates_file: str=None,
            clean_scratch: bool=True, header_str: str=None, **kwargs) -> str:
        """Generate job script contents for the simple sequencer.

        Each job will have a unique shell script generated. The parameters
        passed here are namely used to fill the script templates defined
        in the `sss` module.

        Parameters:
            job_identifier : str
                job ID, nominally the name of the job
            user_script_to_source : str
                full filename (including path) to the user's job script
            timer_name : str
                name of job's timer (used in the .simulation.time.state file)
            sequencing_dates_file : str
                full filename to list of all dates for the simulation
            clean_scratch : bool
                whether to clean the scratch directory (default True)
            header_str : str
                a string that will be written as a comment at the top of
                the job script
            kwargs :
                submit_next : bool
                submit_dependents_on_complete : bool
                output_redirect_path : str (PBSScheduler only)
        """
        # This creates a string (filled template), whose content reflects
        # the core sequencing logic of SSS

        def format_text_block(text):
            return textwrap.dedent(text).strip()

        submit_next = to_shell_bool(kwargs.get('submit_next', True))
        submit_dependents = to_shell_bool(kwargs.get('submit_dependents_on_complete', True))
        output_redirect_path = kwargs.get('output_redirect_path', None)

        # FIXME lazy test
        # if isinstance(self.scheduler, PBSScheduler):
        if (self.scheduler.directive_prefix == "#PBS") and (output_redirect_path is not None):
            # add some trap method to header to help with keeping output files
            # per run (only for previously run file -- and not fool proof!)
            methods = """
            listings_output={PATH}
            function cleanup_previous {{
                jid=$( echo $job_restart_date | sed 's/[ -:_T]//g' )
                [[ -z $jid ]] && jid=$$
                stem="${{listings_output%.*}}"
                ext="${{listings_output##*.}}"
                [[ -e $listings_output ]] && cp $listings_output "$stem"_"$jid"."$ext"
            }}
            trap cleanup_previous EXIT
            """.format(PATH=output_redirect_path)
        else:
            methods = ""

        # format a string for the name of this tool
        pretty_tool_name = ' '.join(re.split(r'(?=[A-Z][a-z])', self.__class__.__name__)).strip()

        if header_str is not None:
            if not (header_str).startswith(SCRIPT_COMMENT_CHAR):
                header_str = f'{SCRIPT_COMMENT_CHAR} {header_str}'
        else:
            header_str = ''

        # create and fill in the script templates

        header = SCRIPT_HEADER_TEMPLATE.format(
            interpreter=SCRIPT_INTERPRETER_DIRECTIVE,
            tool_title=pretty_tool_name,
            date_string=get_date_string(),
            header_comments=header_str,
            methods=SCRIPT_IMSI_ENV_COMMAND + '\n' + format_text_block(methods)
        )

        body = SCRIPT_BODY_TEMPLATE.format(
            runid=self.runid,
            work_dir=self.work_dir,
            job_name=job_identifier,
            timer_name=timer_name,
            timer_tracking_file=self.timer_tracking_file,
            sequencing_dates_file=sequencing_dates_file,
            sequencing_dir=self.sequencer_config_dir,
            listings_dir=self.listings_dir,
            submit_next_bool=submit_next,
            submit_dependents_bool=submit_dependents,
            iss_config_file=os.path.join(self.sequencer_config_dir, '.iss.json'),
            clean_scratch_bool=to_shell_bool(clean_scratch),
            user_script=user_script_to_source,
            run_config_dir=self.run_config_dir,
            sequencing_scratch_dir=self.sequencing_scratch_dir
        )

        footer = SCRIPT_FOOTER                    # dev: as is for now

        # dedent, strip trailing \s, join
        content = []
        for component in [header, body, footer]:
            content.append(format_text_block(component))
        content = '\n\n'.join(content) + '\n'

        return content

    def to_json(self) -> str:
        data = asdict(self)
        data.pop('flow')   # remove
        data['jobs'] = {k: asdict(v) for k, v in self.flow.jobs.items()}
        return json.dumps(data, indent=4)

    def save_to_json(self, file_path: str):
        json_str = self.to_json()
        with open(file_path, 'w') as file:
            file.write(json_str)

    @staticmethod
    def from_json(json_str: str) -> 'IMSISimpleSequencer':
        data = json.loads(json_str)

        # preserve objects and structure to pass back to constructor:
        data['flow'] = SimpleFlow([SimpleJob(**j) for j in data['jobs'].values()])

        # remove so that these are not passed to the IMSISimpleSequencer constructor
        data.pop('jobs')
        ttf = data.pop('timer_tracking_file')
        ld = data.pop('listings_dir')

        # init
        seq = IMSISimpleSequencer(**data)

        # reset / preserve from original input
        seq.timer_tracking_file = ttf
        seq.listings_dir = ld

        return seq

    @staticmethod
    def load_from_json(file_path: str) -> 'IMSISimpleSequencer':
        with open(file_path, 'r') as file:
            json_str = file.read()
        return IMSISimpleSequencer.from_json(json_str)

def iso8601_to_unaware(timestring):
    """Convert iso8601-like datetime string and convert to unaware
    time string (year through second, no sep).
    """
    # lazy type check -- will raise inside function if format isn't valid
    timeinfo = parse_iso8601_like(timestring)
    return "{year}{month}{day}{hour}{minute}{second}".format(**timeinfo)

def is_unaware_datetime_str(timestring):
    """Return 1 if timestring is in unaware datetime timestring format, otherwise return 0."""
    if re.match(UNAWARE_DATETIME_PATTERN, timestring) is None:
        return 0
    else:
        return 1

def to_unaware_datetime_str(timestring):
    """Convert iso8601 timestring to unaware datetime timestring format."""
    if is_unaware_datetime_str(timestring):
        # already unaware format
        return timestring
    else:
        return iso8601_to_unaware(timestring)

def to_unaware_datetime(timestring):
    return datetime.datetime.strptime(to_unaware_datetime_str(timestring), UNAWARE_DATETIME_FORMAT)

def create_status_basename(job_name, start_time, stop_time):
    """Create the filename basename (including extension) the given job
    parameters

    Parameters:
        job_name : str
        start_time, stop_time : str
            datetimes as strings that are iso8601-like OR are in 14-digit
            "unaware" format (`YYYYmmDDHHMMSS`).

    Returns:
        basename of status file: {job_name}_{starttimestr}_{stoptimestr}.status
        where start and stop time strings are in "unaware" format.
    """
    start_info_fmt = to_unaware_datetime_str(start_time)
    stop_info_fmt = to_unaware_datetime_str(stop_time)

    basename = JOB_STATUS_FILENAME_TEMPLATE.format(
        job_name=job_name,
        start_time=start_info_fmt,
        stop_time=stop_info_fmt
        )
    return basename

def set_job_status(job_name, status, start_time, stop_time, tracking_dir=None):
    """set the status of a job in a job state file

    a function that creates a file like
        {RUNID}_{JOB_NAME}_{job_start_date}_{job_stop_date}.state
    that contains:
        the current status, one of allowed iss.ALLOWED_JOB_STATUSES
    """
    if status not in ALLOWED_JOB_STATUSES.values():
        raise ValueError(
            "status '{status}' invalid; status must be one of: {options}".format(status=status, options=', '.join(ALLOWED_JOB_STATUSES.values()))
            )

    if tracking_dir is None:
        tracking_dir = os.getcwd()
    elif not os.path.exists(tracking_dir):
        raise FileNotFoundError(f"can't set status because tracking directory does not exist: {tracking_dir}")

    basename = create_status_basename(job_name, start_time, stop_time)

    # note: always overwrite file
    state_file = os.path.join(tracking_dir, basename)
    with open(state_file, 'w') as f:
        print(status, file=f)

def read_job_status_file_contents(state_file):
    with open(state_file, 'r') as f:
        status_contents = f.readlines()

    # ensure that this file is valid (one line, status only)
    if len(status_contents) == 1:
        status = status_contents[0].strip()
        if status not in ALLOWED_JOB_STATUSES.values():
            raise ValueError(f'invalid status file content: {status}')
    else:
        raise ValueError('invalid status file content')

    return status

def get_job_status(job_name, start_time, stop_time, tracking_dir=None):
    """Get the status of the job (a specific instance).

    The job name, start time, and stop time must all be provided (these form
    the basename of the status file).
    """

    if tracking_dir is None:
        tracking_dir = os.getcwd()

    basename = create_status_basename(job_name, start_time, stop_time)
    state_file = os.path.join(tracking_dir, basename)
    if not os.path.exists(state_file):
        raise FileNotFoundError(f"no tracking file found at: {state_file}")

    status = read_job_status_file_contents(state_file)

    return status

def get_timer_name(job_name, cfg_file) -> str:
    """Get associated timer name for a specific job_name"""
    # NOTE this is to avoid assuming that (job_name == timer_name), though
    # this should be the case.
    seq = IMSISimpleSequencer.load_from_json(cfg_file)
    if job_name not in seq.flow.jobs:
        raise KeyError(f"no job named '{job_name}' in sequencing config {cfg_file}")
    return seq.flow.jobs[job_name].timer_name

def get_job_name(timer_name, cfg_file) -> str:
    """Get the job_name from the given timer_name"""
    # NOTE this is to avoid assuming that (timer_name == job_name), though
    # this should be the case.
    seq = IMSISimpleSequencer.load_from_json(cfg_file)
    job_name = None
    for job, spec in seq.flow.jobs.items():
        if spec.timer_name == timer_name:
            job_name = job
            break
        else:
            pass
    if job_name is None:
        raise KeyError(f"no timer named '{timer_name}' in sequencing config {cfg_file}")
    return job_name

def ready_to_submit(job_name, iss_config_file) -> bool:
    """Determine if the job is ready to submit to the scheduler.

    A job is deemed "ready to submit" when the job status and timer state
    are resolved and the dependency criteria are met.

    Returns:
        True : the job is NOT ready to submit.
        False : the job is ready to submit (state and dependency conditions are met).
    """

    sequencer = IMSISimpleSequencer.load_from_json(iss_config_file)
    listings_dir = os.path.join(sequencer.sequencer_config_dir, "listings")
    timer_tracking_file = sequencer.timer_tracking_file

    job = sequencer.flow.jobs.get(job_name)
    if job is None:
        raise KeyError(f"job '{job_name}' doesnt exist")

    chunk_file = job.sequencing_dates_file
    timer_name = job.timer_name

    chunks = cm.read_chunks(chunk_file)
    current_chunk = cm.get_current_chunk(timer_tracking_file, timer_name)

    try:
        status = get_job_status(job_name, current_chunk['START'], current_chunk['END'], tracking_dir=listings_dir)
    except FileNotFoundError:
        # if the status file doesn't exist yet
        status = None

    # TODO FIXME: note that cm.get_chunk_by_start should be updated, it should
    # NOT not return -1 for index if not found (this is a valid index in python)
    _, idx = cm.get_chunk_by_start(chunks, current_chunk['START'])

    if (status is not None):
        # safety check
        # AND catch if chunk already active/complete
        proceed = False
    elif idx < len(chunks):
        # within list of chunks

        ref_chunk = chunks[idx]

        # use the definitions in the flow (rather than individual jobs)
        # to use convenience objects next
        depends_on = sequencer.flow.job_parents[job_name]

        if not depends_on:
            # no dependencies to check
            proceed = True

        else:
            # check that ALL dependencies are met
            all_dependencies_met = []

            for parent in depends_on:
                p_job = parent.job_name
                p_cond = parent.run_conditional_on
                p_timer_name = sequencer.flow.jobs.get(p_job).timer_name
                p_chunk = cm.get_current_chunk(timer_tracking_file, p_timer_name)

                # get datetimes for dependency and next chunk for the requested condition (START, etc)
                dt_p = to_unaware_datetime(p_chunk[p_cond])
                dt_ref = to_unaware_datetime(ref_chunk[TIME_CHUNK_INDICES[p_cond]])

                # the following ensures that the check is for conditions
                # of the 'most recent' parent job

                if dt_p > dt_ref:
                    # parent is further along than dependent
                    all_dependencies_met.append(True)

                elif dt_ref == dt_p:
                    # if parent and dependent are at the same time then
                    # check that the parent is complete

                    try:
                        p_job_status = get_job_status(
                            p_job, p_chunk["START"], p_chunk["END"], tracking_dir=listings_dir
                            )
                    except FileNotFoundError:
                        # note: this exception is unlikely unless there was
                        # intervention since the timer should not have
                        # advanced on the parent without a status file existing.
                        # since this will cause one condition failure, we can skip
                        # (break from) any other checks
                        all_dependencies_met.append(False)
                        break

                    status_condition_met = True if p_job_status == ALLOWED_JOB_STATUSES['success'] else False
                    all_dependencies_met.append(status_condition_met)

                else:
                    all_dependencies_met.append(False)

            if all(all_dependencies_met):
                # submit this next job
                proceed = True
            else:
                # not able to submit this next job
                proceed = False

    else:
        # safety - ran out of chunks
        proceed = False

    return proceed

def is_first_iteration(job_name, timer_tracking_file, chunk_file,
                       timer_name=None, listings_dir=None):
    # resolve the state of the job status between the timer tracking info
    # and the job's state
    if timer_name is None:
        timer_name = job_name

    is_first_chunk = cm.check_if_first_chunk(timer_name, chunk_file, timer_tracking_file)

    if is_first_chunk:
        timeinfo = cm.get_current_chunk(timer_tracking_file, timer_name)
        start_fmt = iso8601_to_unaware(timeinfo['START'])
        stop_fmt = iso8601_to_unaware(timeinfo['END'])

        try:
            status = get_job_status(job_name, start_fmt, stop_fmt, tracking_dir=listings_dir)
        except FileNotFoundError:
            status = None

        if status is None or (status == ALLOWED_JOB_STATUSES['fail']):
            # status didn't exist or previously failed (treated equivalently)
            first_iter = True
        else:
            first_iter = False

    else:
        first_iter = False

    return first_iter

def get_job_status_summary(job_name, tracking_dir=None):
    """Return a dict of status file names and file contents (the status),
    and a list of files that could not be read (errors).

    Paramters:
        job_name : job name
        tracking_dir : path to listings directory (location of .status files)

    Returns:
        statuses, errors
    """
    if tracking_dir is None:
        tracking_dir = os.getcwd()

    status_filelist = sorted(glob(os.path.join(tracking_dir, f"{job_name}*.status")))

    stats = {}
    error_files = []

    if len(status_filelist) == 0:
        pass
    else:
        for f in status_filelist:
            try:
                status = read_job_status_file_contents(f)
            except ValueError:
                error_files.append(f)
                continue
            stats[f] = status

    return stats, error_files

def get_job_status_summary_message(job_name, tracking_dir=None):
    """Return a formatted string of job status summary information.

    The string includes:
        - A list of status file names (basename, relative to `tracking_dir`)
        and the status (file content)
        - If applicable, a list of files that could not be interpreted.
    """
    if tracking_dir is None:
        tracking_dir = os.getcwd()

    statuses, error_files = get_job_status_summary(job_name, tracking_dir=tracking_dir)

    header_strs = [f'job status for: {job_name}', f'listings in: {tracking_dir}']
    output_strs = header_strs

    if statuses:
        status_strs = []
        for f,s in statuses.items():
            f_display = os.path.basename(f).replace('.status', '')
            status_strs.append("{}  {}".format(f_display, s))
        output_strs += status_strs
    else:
        output_strs += ['no status files found.']

    output_msgs = ['\n'.join(output_strs)]

    if error_files:
        error_str = "\n".join(error_files)
        output_msgs += ["\n".join([
            'job status could not be read from the following file(s):',
            f'{error_str}'
            ])
        ]
    msg = '\n\n'.join(output_msgs)
    return msg

def _get_iss_cfg_file(cfg_file):
    # returns path to iss config file (json), in current location
    # if not specified
    if cfg_file is None:
        cfg_file = os.path.join(os.getcwd(), '.iss.json')
    return cfg_file

def which_scheduler(cfg_file=None):
    """Return the scheduler name for the iss run."""
    cfg_file = _get_iss_cfg_file(cfg_file)
    sequencer = IMSISimpleSequencer.load_from_json(cfg_file)
    n = sequencer.scheduler.get('name', None)
    if n is None:
        n = sequencer.scheduler.get('directive_prefix')
    return n

def which_submit(cfg_file=None):
    """Return the scheduler submission command for the iss run."""
    cfg_file = _get_iss_cfg_file(cfg_file)
    sequencer = IMSISimpleSequencer.load_from_json(cfg_file)
    n = sequencer.scheduler.get('submission_command', None)
    if n is None:
        raise AttributeError('no command specified')
    return n

def get_job_chunk_dict(cfg_file):
    """Return the list of time chunks for a specific job for the iss run."""
    cfg = _get_iss_cfg_file(cfg_file)
    sequencer = IMSISimpleSequencer.load_from_json(cfg)

    jobs = sequencer.flow.jobs.keys()
    job_chunks = {}
    for j in jobs:
        chunk_file = sequencer.flow.jobs[j].sequencing_dates_file
        job_chunks[j] = cm.read_chunks(chunk_file)
    return job_chunks

def which_jobs(cfg_file=None):
    """Return a formatted string of the time chunks for each jobs
    of the iss run.
    """
    job_chunks = get_job_chunk_dict(cfg_file)

    LINE_TEMPLATE = "{:>7}{:>22}{:>22}{:>22}"
    header = list(TIME_CHUNK_INDICES.keys())

    msgs = ["jobs:"]
    for j,c in job_chunks.items():
        msgs.append("{job} (n = {n})".format(job=j, n=len(c)))
        msgs.append(LINE_TEMPLATE.format(*([''] + header)))
        msgs.append(('\n'.join([LINE_TEMPLATE.format(*([i+1] + row)) for i,row in enumerate(c)])))

    return '\n'.join(msgs)



"""
utils
=====

Utility functions used in imsi, largely for parsing json,
and updating, searching and modifying nested python dicts.

"""
import traceback
import yaml
import json5
import collections
import fnmatch
import os
import re
from collections import Counter
from pathlib import Path
from omegaconf import OmegaConf

def load_config_file(config_file):
    """Reads a config file using the appropriate
    function for the file type (supports yaml or json)"""
    if config_file.suffix == '.yaml':
        return OmegaConf.load(config_file)
    elif config_file.suffix in [".json", ".jsonc"]:
        return load_json(config_file)
    else:
        raise ValueError(f"Unsupported file type: {config_file}")

def load_json(config_file):
    """Reads a json config file"""
    with open(config_file) as f:
        # Note order can be important so use ordered dicts
        config = json5.load(f, object_pairs_hook={})
    return config


def update(d,u, verbose=False):
    """
    Recursively update a dictionary, d, with an update, u.
    If an item appears in only one dictionary, it is included in the result.

    Parameters:
      d : dict to update
      u : dict of updates

    Returns:
    --------
      d : updated dict
    """
    #Based on:
    #https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
    #Note this is the workhorse of all imsi inheritance behaviour.
    for k, v in u.items():
        if verbose:
            print(f"Updating {k}={v}")
        if isinstance(v, collections.abc.Mapping):
            d[k] = update(d.get(k, {}), v)
        else:
            d[k] = v
    return d

def flatten(d, parent_key='', sep='_'):
    """Flatten a nested dict, using sep to join keys from levels in order"""
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.abc.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)

def recursive_lookup(key, d):
    """Find a unique key in a nested dict. Will not be graceful if there are duplicates!"""
    for k, v in d.items():
        if k == key:
            yield v
        elif isinstance(v, collections.abc.MutableMapping):
            for result in recursive_lookup(key, v):
                yield result

def combine_yaml_configs(rootpath):
    """combine all YAML files found recursively under rootpath into one dictionary.
    This effectively provides a dictionary-database to use.

    Parameters:
       rootpath : str
          The path to recursively search for input files ending in .yaml
          Normally this is the path to the imsi-config directory.

    Outputs:
       config: dict
          A dictionary of the contents of the json or combined json files found under rootpath.
    """
    # Note: obviously combine_configs should be generalized to arbitrary data format, just doing for now...

    config = {}
    for source_subpath in Path(rootpath).rglob("*.yaml"):
        try:
            temp_config = OmegaConf.load(source_subpath)
            temp_config = OmegaConf.to_container(temp_config, resolve=True)
            config = update(config, temp_config)
        except yaml.scanner.ScannerError:
            traceback.print_exc(limit=1)
            exit(1)

    return config

def combine_json_configs(rootpath):
    """combine all json files found recursively under rootpath into one dictionary.
       This effectively provides a dictionary-database to use.

       Parameters:
          rootpath : str
             The path to recursively search for input files ending in .json/.jsonc.
             Normally this is the path to the imsi-config directory.

       Outputs:
          config: dict
             A dictionary of the contents of the json or combined json files found under rootpath.
    """
    files = []
    config = {}
    patterns = ['*.json', '*.jsonc']
    for root, dir, filenames in os.walk(rootpath):
        for pattern in patterns:
            for f in fnmatch.filter(filenames, pattern):
                files.append(os.path.join(root, f))
                temp_dict = load_json(os.path.join(root, f))
                config = update(config, temp_dict)
    return config

def resolve_inheritance(configs, selected_config, config_hierarchy):
    """
    Resolve the inheritance tree through recursion and exclusion of duplicates.
    Ensure ancestors precede descendents.

    Parameters:
       configs: dict
          dictionary of all possible configurations, typically from an imsi_database.
       selected_config: str
          The configuration to choose
       config_hierarchy: list
          Typically not user specified, but used in the recursive function call to
          layer configurations in the correct order of inheritance.
    """

    config = configs[selected_config]

    if 'inherits_from' in config and config['inherits_from']:
        parent_configs = config['inherits_from']
        if isinstance(parent_configs, str):
            parent_configs = [parent_configs]

        for parent_config in reversed(parent_configs):
            if parent_config in configs.keys():
                config_hierarchy = resolve_inheritance(configs=configs, selected_config=parent_config, config_hierarchy=config_hierarchy)

            else:
                raise NameError(f'{parent_config} not found. Check inheritance in {selected_config}')

    if config not in config_hierarchy:
        config_hierarchy.append(config)

    return config_hierarchy

def parse_config_inheritance(configs, selected_config):
    """
    Parse a configuration, inheriting attributes from all "parents".

    Parameters:
       configs: dict
          dictionary of all possible configurations, typically from an imsi_database.
       selected_config: str
          The configuration to choose
       config_hierarchy: list
          Typically not user specified, but used in the recursive function call to
          layer configurations in the correct order of inheritance.
    """

    config = configs[selected_config]

    config_hierarchy = []
    config_hierarchy = resolve_inheritance(configs, selected_config, config_hierarchy)

    if len(config_hierarchy) > 1:
        config = {}

        for child in config_hierarchy:
            config = update(config, child)

    return config

def parse_config_inheritance_org(configs, selected_config, config_hierachy=None):
    """
    Parse a configuration recursively inheriting attributes from all "parents".

    Parameters:
       configs: dict
          dictionary of all possible configurations, typically from an imsi_database.
       selected_config: str
          The configuration to choose
       config_hierachy: dict
          Typically not user specified, but used in the recursive function call to
          layer configurations in the correct order of inheritance.
    """
    # Curiously, assigning config_hierachy to an empty default
    # leads to persistent values leaking across function calls. -> probably due to lists being mutable.
    if config_hierachy is None:
        config_hierachy = []
    config = configs[selected_config]
    config_hierachy.append(config)
    if 'inherits_from' in config and config['inherits_from']:
        parent_config = config['inherits_from']
        if parent_config in configs.keys():
            config = parse_config_inheritance(configs=configs, selected_config=parent_config, config_hierachy=config_hierachy)
        else:
            raise NameError(f'{parent_config} not found. Check inheritance in {selected_config}')
    else:
        if len(config_hierachy)>1:
            config = {}
            config_hierachy.reverse()
            for child in config_hierachy:
                config = update(config, child)
    return config

def parse_var(s):
    """
    Parse a key, value pair, separated by '='
    That's the reverse of ShellArgs.
    On the command line (argparse) a declaration will typically look like::

        foo=hello

    or::

        foo="hello world"

    Courtesy https://stackoverflow.com/questions/27146262/create-variable-key-value-pairs-with-argparse-python
    """
    items = s.split('=')
    key = items[0].strip() # we remove blanks around keys, as is logical
    if len(items) > 1:
        # rejoin the rest:
        value = '='.join(items[1:])
    return (key, value)


def parse_vars(items, none_as_str=True):
    """
    Parse a series of key-value pairs and return a dictionary
    Courtesy https://stackoverflow.com/questions/27146262/create-variable-key-value-pairs-with-argparse-python

    """
    d = {}

    if items:
        for item in items:
            key, value = parse_var(item)
            if not none_as_str and value == "None":
                value = None
            d[key] = value
    return d

# ChatGPT beauty
def replace_variables(obj, inputs):
    if isinstance(obj, dict):
        # If obj is a dictionary, iterate through its key-value pairs
        for key, value in obj.items():
            obj[key] = replace_variables(value, inputs)
        return obj
    elif isinstance(obj, list):
        # If obj is a list, iterate through its elements
        return [replace_variables(item, inputs) for item in obj]
    elif isinstance(obj, str):
        # If obj is a string, perform variable replacement
        return replace_variables_in_string(obj, inputs)
    else:
        # For other data types, return the object as is
        return obj

def replace_variables_in_string(s, inputs):
    # Define a function to replace variables in a string
    def replace(match):
        variable_name = match.group(1)
        if variable_name in inputs:
            return str(inputs[variable_name])  # Convert the value to a string
        else:
            return match.group(0)  # If variable not found, keep it as is


    # Use regular expression to find and replace variables in the string
    pattern = r'{{(.*?)}}'
    return re.sub(pattern, replace, s)

def replace_curlies_in_dict(input_dict, replacement_dict):
    pattern = re.compile(r"{{(.*?)}}")

    def replace_in_string(s, repl_dict):
        matches = pattern.findall(s)
        for match in matches:
            value = deep_get(repl_dict, match)
            if value is not None:
                s = s.replace(f"{{{{{match}}}}}", str(value))
        return s

    def deep_get(d, key):
        if isinstance(d, dict):
            if key in d:
                return d[key]
            for k, v in d.items():
                result = deep_get(v, key)
                if result is not None:
                    return result
        elif isinstance(d, list):
            for item in d:
                result = deep_get(item, key)
                if result is not None:
                    return result
        return None

    def recursive_replace(d, repl_dict):
        if isinstance(d, dict):
            return {k: recursive_replace(v, repl_dict) for k, v in d.items()}
        elif isinstance(d, list):
            return [recursive_replace(i, repl_dict) for i in d]
        elif isinstance(d, str):
            return replace_in_string(d, repl_dict)
        else:
            return d

    def find_referenced_keys(d, pattern):
        referenced_keys = set()
        if isinstance(d, dict):
            for k, v in d.items():
                if isinstance(v, str):
                    matches = pattern.findall(v)
                    referenced_keys.update(matches)
                else:
                    referenced_keys.update(find_referenced_keys(v, pattern))
        elif isinstance(d, list):
            for item in d:
                referenced_keys.update(find_referenced_keys(item, pattern))
        return referenced_keys

    def check_for_duplicate_keys(replacement_dict, referenced_keys):
        key_counter = Counter()
        def count_keys(d):
            if isinstance(d, dict):
                for k, v in d.items():
                    if k in referenced_keys:
                        key_counter[k] += 1
                    count_keys(v)
            elif isinstance(d, list):
                for item in d:
                    count_keys(item)

        count_keys(replacement_dict)

        duplicates = [k for k, count in key_counter.items() if count > 1]
        if duplicates:
            raise ValueError(f"Duplicate keys found in replacement_dict for referenced variables: {', '.join(duplicates)}")

    referenced_keys = find_referenced_keys(input_dict, pattern)
    check_for_duplicate_keys(replacement_dict, referenced_keys)

    return recursive_replace(input_dict, replacement_dict)



import pytest
from pydantic import ValidationError
from imsi.config_manager.schema.setup_params import SetupParams

# TODO this is currently a stub, to be refined and expanded

@pytest.fixture
def valid_setup_params_data():
    """Fixture to provide valid setup param data."""
    return {
        'runid': 'abc-123',
        'model_name': 'test_model',
        'experiment_name': 'test_exp',
        'machine_name': 'test_machine',
        'compiler_name': 'test_compiler',
        'sequencer_name': 'test_seq',
        'flow_name': 'test_flow',
        'postproc_profile': 'test_postproc',
        'work_dir': '/path/to/work_dir',
        'run_config_path': '/path/to/work_dir/config',
        'source_repo': '/path/to/work_dir/source_repo',
        'fetch_method': 'copy',
        'source_version': None,
        'source_path': '/path/to/work_dir/src',
        'imsi_config_path': '/path/to/work_dir/src/imsi-config',
        'imsi_venv': 'imsi_venv_path'
    }


def test_init(valid_setup_params_data):
    SetupParams(**valid_setup_params_data)


def test_required_fields(valid_setup_params_data):
    """Ensure some fields can be None."""
    valid_setup_params_data['machine_name'] = None
    valid_setup_params_data['compiler_name'] = None
    valid_setup_params_data['sequencer_name'] = None
    valid_setup_params_data['flow_name'] = None

    sp = SetupParams(**valid_setup_params_data)

    assert sp.machine_name is None
    assert sp.compiler_name is None
    assert sp.sequencer_name is None
    assert sp.flow_name is None


def test_invalid_path_setup(valid_setup_params_data):
    valid_setup_params_data['source_path'] = '/fake/path/imsi-config'
    with pytest.raises(ValidationError):
        SetupParams(**valid_setup_params_data)



import os
import shutil
from pathlib import Path
import warnings

import yaml

from imsi.utils.dict_tools import update
from imsi.config_manager.config_manager import Configuration, ConfigManager


def _get_config_filename(serialized=False) -> str:
    # determine the filename from cwd or WRK_DIR for the user-facing or
    # serialized config file

    def _get_basename(dirname, serialized):
        if serialized:
            stembase = f".imsi/.imsi_configuration_{dirname}.pickle"
        else:
            stembase = f"imsi_configuration_{dirname}.yaml"
        return stembase

    # First look in the PWD
    current_dir = './'   # potential work_dir, to check
    wrk_dir = current_dir  # TODO
    current_dir_resolved = Path(wrk_dir).resolve()
    current_dir_name = current_dir_resolved.name

    config_filename_base = _get_basename(current_dir_name, serialized)
    guessed = os.path.join(current_dir_resolved, config_filename_base)

    if not os.path.exists(guessed):
        # now assume that if the config file didnt match the name of the
        # cwd, then maybe WRK_DIR is set instead
        wrk_dir = os.getenv("WRK_DIR")

        cfg_type_msg = "serialized configuration" if serialized else "configuration"

        if wrk_dir:
            current_dir_name = Path(wrk_dir).name
            config_filename_base = _get_basename(current_dir_name, serialized)
            guessed = os.path.join(wrk_dir, config_filename_base)
            if os.path.exists(guessed):
                warnings.warn(f"\nNB: Using configuration defined by WRK_DIR at {guessed}\n")
            else:
                raise FileNotFoundError(
                    f"Cannot find the imsi {cfg_type_msg} file expected: "
                    f"{guessed}"
                    )
        else:
            raise FileNotFoundError(
                f"Cannot find the imsi {cfg_type_msg} file expected: {guessed} "
                "You must be in a valid imsi setup directory or WRK_DIR must be defined"
                )

    # by here, the guessed file is the confirmed imsi configuration file
    return guessed


def apply_options_overrides(config_dictionary: dict, options: list) -> dict:
    """Merge YAML option files into config_dictionary.

    Each item in options must be a single-key dict mapping a subdirectory name to
    a filename (without suffix). Example: [{"some_dir": "option_name"}, ...].

    If any loaded YAML defines keys whose full key-paths overlap with existing
    keys in config_dictionary, a warning is emitted listing the overlapping paths.
    """
    options_root = Path(config_dictionary["setup_params"]["work_dir"]) / "src" / "imsi-config"

    for option in options:
        parent_dir, filename = next(iter(option.items()))
        option_file = (options_root / parent_dir / filename).with_suffix(".yaml")

        if not option_file.exists():
            raise FileNotFoundError(f"Option file not found: {option_file}")

        with option_file.open() as stream:
            cfg = yaml.safe_load(stream) or {}

        config_dictionary = update(config_dictionary, cfg)

    return config_dictionary


def load_run_config(serialized=True):
    # relative to cwd/pwd
    config_file = _get_config_filename(serialized=serialized)
    config_manager = ConfigManager()
    if serialized:
        configuration = config_manager.load_state(config_file)
    else:
        configuration = config_manager.load_configuration(config_file)
    return configuration


def freeze_run_configuration(configuration: Configuration):
    """Copies the current imsi configuration file to the configuration folder
    as a dotfile.

    By convention:
        source:       {work_dir}/imsi_configuration_{runid}.yaml
        destination:  {run_config_path}/.imsi_configuration_{runid}.yaml
    """
    wrk_dir = configuration.setup_params.work_dir
    runid = configuration.setup_params.runid
    run_config_path = configuration.setup_params.run_config_path

    basename = f"imsi_configuration_{runid}.yaml"
    src = os.path.join(wrk_dir, basename)
    if not os.path.exists(src):
        raise FileNotFoundError(f"ERROR: missing imsi configuration file {src}")
    dst = os.path.join(run_config_path, f".{basename}")
    shutil.copy2(src, dst)

def save_setup_configuration(configuration: Configuration, save_config : bool = True):
    """Save the resolved imsi configuration state to files.
    A file will be written to the working directory (work_dir) and a copy
    will be made under the config directory (run_config_path) as a hidden
    file.

    By convention the basename of the file is imsi_configuration_{runid}.yaml
    """
    wrk_dir = configuration.setup_params.work_dir
    runid = configuration.setup_params.runid
    config_manager = ConfigManager()

    # create a .imsi folder if it does not exist
    Path(wrk_dir, ".imsi").mkdir(exist_ok=True)
    hidden_imsi_dir = Path(wrk_dir, ".imsi")

    config_manager.save_state(
        configuration,
        filepath=Path(wrk_dir)
        / hidden_imsi_dir
        / f".imsi_configuration_{runid}.pickle",
    )
    if save_config:
        config_manager.save_configuration(
            configuration,
            filepath=Path(wrk_dir) / f"imsi_configuration_{runid}.yaml",
        )



from pydantic import BaseModel, ConfigDict, Field
from typing import List, Optional, Union


class Model(BaseModel):
    """Model configuration Pydantic model"""

    model_config = ConfigDict(extra='allow')
    name: str = Field(..., description='Model name')
    source_id: str
    variant_id: str
    short_name: str
    description: str
    postproc_profile: str
    inherits_from: Optional[Union[List[str], str]] = None
    Scientifically_validated: Optional[bool] = None
    repository_tag: Optional[str] = None
    uxxx: Optional[str] = None
    prefix: Optional[str] = None
    model_filename_prefix: Optional[str] = None
    model_rs_filename_prefix: Optional[str] = None
    flow: Optional[str] = None



from imsi.cli.entry import cli
import shutil
# import pytest
from pathlib import Path
from click.testing import CliRunner
import os
import subprocess
from imsi.tools.time_manager.time_manager import sim_time_factory
from imsi.shell_interface.shell_timing_vars import generate_timers_config
from imsi.utils.general import write_shell_script

#WIP: testing ground for time-manager and chunk-manager routines
#Need to implement testing protocol (pytest?)
#and test a number of configurations.
#Also, diff output files and cleanup afterwards.

#At present, this will run each of the three chunk manager methods:
# loop index, tracker file, and iso8601 date
# for a two year segment.
# The index method and tracker method below
# attempt to run a third (out of bounds) year and we see the result:
# index method will raise an error
# tracker method will duplicate the timers for 2nd year

# Test settings
run_dates = {
        "run_start_time": "6000-01-01T00:00:00",
        "run_stop_time": "6010-12-31T23:59:59",
        "run_segment_start_time": "6000-01-01T00:00:00",
        "run_segment_stop_time": "6001-12-31T23:59:59",
        "model_chunk_size": "P12MS",
        "model_internal_chunk_size": "P12MS",
        "postproc_chunk_size": "P12MS"
}
prefix = "job"
output_file = "job_time_params"

output_path = Path('chunk_manager_2yr_test')
output_path.mkdir(exist_ok=True)

#First output the dates file using the time manager.
#This file is still in use by ISS/SSS though will
#not be used anymore by Maestro. We need to do some
#more careful work to see if we can remove the need
#for dates files in ISS/SSS
run_config_path = output_path / 'dates_files'
run_config_path.mkdir(exist_ok=True)
dates_list = run_config_path / 'model_submission_job_start-stop_dates'
if dates_list.exists():
    dates_list.unlink()
write_shell_script(os.path.join(str(run_config_path), 'model_submission_job_start-stop_dates'),
                  generate_timers_config(run_dates, 'model_submission_job'))

###################################### index method
#test loop index method (maestro and model inner loop)
#put in nested directory
new_dir = output_path / "chunk_index_method"
new_dir.mkdir(exist_ok=True)

for loop_index in range(1,4):
    cmd = f"imsi chunk-manager create-time-env-file from-index --start-time={run_dates['run_segment_start_time']} " + \
        f"--stop-time={run_dates['run_segment_stop_time']} --chunk-size={run_dates['model_chunk_size']} " + \
        f"--loop-index={loop_index} --prefix={prefix} --output-file={str(new_dir)}/{output_file}_loop_{loop_index}"
    proc = subprocess.run(cmd, shell=True)

###################################### tracker method
#test tracker file routines (iss/sss)
new_dir = output_path / "chunk_tracker_method"
new_dir.mkdir(exist_ok=True)

# test init for creation of tracker file
chunk_file = Path(run_config_path) / 'model_submission_job_start-stop_dates'
tracker = new_dir / '.simulation.time.state'
if tracker.exists():
    tracker.unlink()
timer = 'model-submission'
cmd = f"imsi chunk-manager init --chunk-file={str(chunk_file)} --tracking-file={str(tracker)} --timer-name={timer}"
proc = subprocess.run(cmd, shell=True)

for loop_index in range(1,4):
    cmd = f"imsi chunk-manager create-time-env-file from-tracking-file --tracking-file={tracker} " + \
        f"--timer-name={timer} --prefix={prefix} --output-file={str(new_dir)}/{output_file}_loop_{loop_index}"
    proc = subprocess.run(cmd, shell=True)

    #increment to the next timer
    cmd = f"imsi chunk-manager increment --chunk-file={str(chunk_file)} --tracking-file={tracker} --timer-name={timer}"
    proc = subprocess.run(cmd, shell=True)


###################################### isodate method
#test isodate method (cylc)
new_dir = output_path / "chunk_iso8601_method"
new_dir.mkdir(exist_ok=True)

simTime =  sim_time_factory(run_dates, 'model_submission_job')

loop_index = 1
for chunk in simTime.ChunkIndexStart:
    cmd = f"imsi chunk-manager create-time-env-file from-date --isostr={chunk.isoformat()} --chunk-delta={simTime.chunk_size_str} " + \
          f"--prefix={prefix} --output-file={str(new_dir)}/{output_file}_loop_{loop_index}"
    proc = subprocess.run(cmd, shell=True)
    loop_index += 1


'''
General utility functions for creating the "shell interface" (config directory) and associated files.
'''
from typing import List
import os
import shutil

from imsi.utils.git_tools import init_repo, clear_repo
from imsi.utils.general import delete_or_abort, yes_or_no

def create_config_dirs(run_config_path: str, components: List[str], track=True, force=False):
    """Create the run config dir and subdirectories for each component."""

    if os.path.exists(run_config_path):
        if force:
            print(f"**WARNING**: config dir exists, overwriting at {run_config_path}")
        else:
            yes_or_no(f'{run_config_path} already exists, do you want to replace it?')

        if track:
            clear_repo(run_config_path)
        else:
            shutil.rmtree(run_config_path)

    os.makedirs(run_config_path, exist_ok=True)
    for component in components:
        os.makedirs(os.path.join(run_config_path, component), exist_ok=True)

    if track:
        branch = os.path.basename(os.path.dirname(run_config_path))  # convention
        init_repo(path=run_config_path, branch=branch)


if __name__ == "__main__":
    # Only for original dev testing. See shell_interface_manager for more up to date tests.
    pass



import pytest
from click.testing import CliRunner

from imsi.tools.ensemble.ensemble_cli import ensemble


@pytest.fixture
def dummy_config(tmp_path):
    """Creates a dummy config.yaml file that won't be parsed but will bypass file existence checks."""
    config = tmp_path / "config.yaml"
    config.write_text(
        """
        ensemble_level:
          user: dummy
          run_directory: .
        member_level: {}
        """)  # Content isn't used
    return config


@pytest.fixture
def runner():
    return CliRunner()


@pytest.mark.parametrize(
    ("command", "expected_message"),
    [
        (None, "Attempting to use configuration file"),
        ("--help", "Commands to manage ensembles of imsi runs"),
    ],
    ids=["no_command", "help_command"]
)
def test_ensemble_base_help(runner, dummy_config, command, expected_message):
    """Test base CLI group with no subcommand or --help."""
    args = ["--config-path", str(dummy_config)]
    if command:
        args.append(command)
    result = runner.invoke(ensemble, args)
    assert result.exit_code == 0
    assert expected_message in result.output


@pytest.mark.parametrize("subcommand", ["setup", "config", "save-restarts", "build", "submit", "status"])
def test_subcommands_help(runner, dummy_config, subcommand):
    """Each subcommand should show help and be callable."""
    result = runner.invoke(ensemble, ["--config-path", str(dummy_config), subcommand, "--help"])
    assert result.exit_code == 0
    assert subcommand in result.output
